from typing import Any, Optional

from sirius_sdk.agent.connections import AgentRPC
from sirius_sdk.agent.wallet.abstract.crypto import AbstractCrypto


class CryptoProxy(AbstractCrypto):

    def __init__(self, rpc: AgentRPC):
        self.__rpc = rpc

    async def create_key(self, seed: str = None, crypto_type: str = None) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/create_key',
            params=dict(seed=seed, crypto_type=crypto_type)
        )

    async def set_key_metadata(self, verkey: str, metadata: dict) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/set_key_metadata',
            params=dict(verkey=verkey, metadata=metadata)
        )

    async def get_key_metadata(self, verkey: str) -> Optional[dict]:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_key_metadata',
            params=dict(verkey=verkey)
        )

    async def crypto_sign(self, signer_vk: str, msg: bytes) -> bytes:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/crypto_sign',
            params=dict(signer_vk=signer_vk, msg=msg)
        )

    async def crypto_verify(self, signer_vk: str, msg: bytes, signature: bytes) -> bool:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/crypto_verify',
            params=dict(signer_vk=signer_vk, msg=msg, signature=signature)
        )

    async def anon_crypt(self, recipient_vk: str, msg: bytes) -> bytes:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/anon_crypt',
            params=dict(recipient_vk=recipient_vk, msg=msg)
        )

    async def anon_decrypt(self, recipient_vk: str, encrypted_msg: bytes) -> bytes:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/anon_decrypt',
            params=dict(recipient_vk=recipient_vk, encrypted_msg=encrypted_msg)
        )

    async def pack_message(self, message: Any, recipient_verkeys: list, sender_verkey: str = None) -> bytes:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/pack_message',
            params=dict(message=message, recipient_verkeys=recipient_verkeys, sender_verkey=sender_verkey)
        )

    async def unpack_message(self, jwe: bytes) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/unpack_message',
            params=dict(jwe=jwe)
        )



from sirius_sdk.agent.connections import AgentRPC
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache, PurgeOptions, CacheOptions


class CacheProxy(AbstractCache):

    def __init__(self, rpc: AgentRPC):
        self.__rpc = rpc

    async def get_schema(self, pool_name: str, submitter_did: str, id_: str, options: CacheOptions) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_schema',
            params=dict(pool_name=pool_name, submitter_did=submitter_did, id_=id_, options=options)
        )

    async def get_cred_def(self, pool_name: str, submitter_did: str, id_: str, options: CacheOptions) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_cred_def',
            params=dict(pool_name=pool_name, submitter_did=submitter_did, id_=id_, options=options)
        )

    async def purge_schema_cache(self, options: PurgeOptions) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/purge_schema_cache',
            params=dict(options=options)
        )

    async def purge_cred_def_cache(self, options: PurgeOptions) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/purge_cred_def_cache',
            params=dict(options=options)
        )



import uuid
import json
import base64
import logging
import datetime
from typing import Any, Optional

from sirius_sdk.errors.exceptions import *
from sirius_sdk.errors.indy_exceptions import *
from sirius_sdk.rpc.tunnel import AddressedTunnel

MSG_TYPE = 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/future'


class Future:
    """Futures and Promises pattern.
    (http://dist-prog-book.com/chapter/2/futures.html)


    Server point has internal communication schemas and communication addresses for
    Aries super-protocol/sub-protocol behaviour
    (https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0003-protocols).

    Future hide communication addresses specifics of server-side service (cloud agent) and pairwise configuration
    of communication between sdk-side and agent-side logic, allowing to take attention on
    response awaiting routines.
    """

    def __init__(self, tunnel: AddressedTunnel, expiration_time: datetime.datetime = None):
        """
        :param tunnel: communication tunnel for server-side cloud agent
        :param expiration_time: time of response expiration
        """
        self.__id = uuid.uuid4().hex
        self._value = None
        self.__read_ok = False
        self.__tunnel = tunnel
        self.__exception = None
        self.expiration_time = expiration_time

    @property
    def promise(self):
        """
        Promise info builder

        :return: serialized promise dump
        """
        return {
            'id': self.__id,
            'channel_address': self.__tunnel.address,
            'expiration_stamp': self.expiration_time.timestamp() if self.expiration_time else None
        }

    async def wait(self, timeout: int=None) -> bool:
        """Wait for response

        :param timeout: waiting timeout in seconds
        :return: True/False
        """
        if self.__read_ok:
            return True
        try:
            if timeout == 0:
                return False
            if self.expiration_time:
                expires_time = self.expiration_time
            elif timeout:
                expires_time = datetime.datetime.now() + datetime.timedelta(seconds=timeout)
            else:
                expires_time = datetime.datetime.now() + datetime.timedelta(days=365)
            while datetime.datetime.now() < expires_time:
                timedelta = expires_time - datetime.datetime.now()
                timeout = max(timedelta.seconds, 0)
                payload = await self.__tunnel.receive(timeout)
                if (payload.get('@type') == MSG_TYPE) and (payload.get('~thread', {}).get('thid', None) == self.__id):
                    exception = payload['exception']
                    if exception:
                        self.__exception = exception
                    else:
                        value = payload['value']
                        if payload['is_tuple']:
                            self._value = tuple(value)
                        elif payload['is_bytes']:
                            self._value = base64.b64decode(value.encode('ascii'))
                        else:
                            self._value = value
                    self.__read_ok = True
                    return True
                else:
                    logging.warning(
                        'Unexpected payload \n' + json.dumps(payload, indent=2, sort_keys=True) +
                        '\n Expected id: "%s"' % self.__id
                    )
            return False
        except SiriusTimeoutIO:
            return False

    def get_value(self) -> Any:
        """Get response value.

        :return: value
        :raises:
           - SiriusPendingOperation: response was not received yet. Call walt(0) to safely check value persists.
        """
        if self.__read_ok is False:
            raise SiriusPendingOperation()
        return self._value

    def has_exception(self) -> bool:
        """Check if response was interrupted with exception

        :return: True if request have done with exception
        :raises:
           - SiriusPendingOperation: response was not received yet. Call walt(0) to safely check value persists.
        """
        if self.__read_ok is False:
            raise SiriusPendingOperation()
        return self.__exception is not None

    @property
    def exception(self) -> Optional[Exception]:
        """Get exception that have interrupted response routine on server-side.

        :return: Exception instance or None if it does not exists
        """
        if self.has_exception():
            if self.__exception['indy']:
                indy_exc = self.__exception['indy']
                exc_class = errorcode_to_exception(errorcode=indy_exc['error_code'])
                exc = exc_class(
                    error_code=indy_exc['error_code'],
                    error_details=dict(message=indy_exc['message'], indy_backtrace=None)
                )
                return exc
            else:
                return SiriusPromiseContextException(
                    class_name=self.__exception['class_name'], printable=self.__exception['printable']
                )
        else:
            return None

    def raise_exception(self):
        """Raise exception if exists

        :raises:
           - SiriusValueEmpty: raises if exception is empty
        """
        if self.has_exception():
            raise self.exception
        else:
            raise SiriusValueEmpty()



from typing import Optional, List, Any

from sirius_sdk.agent.wallet.abstract.pairwise import AbstractPairwise
from sirius_sdk.agent.connections import AgentRPC


class PairwiseProxy(AbstractPairwise):

    def __init__(self, rpc: AgentRPC):
        self.__rpc = rpc

    async def is_pairwise_exists(self, their_did: str) -> bool:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/is_pairwise_exists',
            params=dict(their_did=their_did)
        )

    async def create_pairwise(self, their_did: str, my_did: str, metadata: dict = None, tags: dict = None) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/create_pairwise',
            params=dict(their_did=their_did, my_did=my_did, metadata=metadata, tags=tags)
        )

    async def list_pairwise(self) -> List[Any]:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/list_pairwise'
        )

    async def get_pairwise(self, their_did: str) -> Optional[dict]:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_pairwise',
            params=dict(their_did=their_did)
        )

    async def set_pairwise_metadata(self, their_did: str, metadata: dict = None, tags: dict = None) -> None:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/set_pairwise_metadata',
            params=dict(their_did=their_did, metadata=metadata, tags=tags)
        )

    async def search(self, tags: dict, limit: int = None) -> (List[dict], int):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/search_pairwise',
            params=dict(tags=tags, limit=limit)
        )



import json
import logging

from sirius_sdk.agent.pairwise import Pairwise, TheirEndpoint
from sirius_sdk.agent.agent import Endpoint
from sirius_sdk.agent.sm import AbstractStateMachine
from sirius_sdk.agent.aries_rfc.feature_0015_acks import Ack, Status
from sirius_sdk.agent.aries_rfc.feature_0048_trust_ping import Ping
from sirius_sdk.agent.aries_rfc.feature_0160_connection_protocol.messages import *


# Problem codes
REQUEST_NOT_ACCEPTED = "request_not_accepted"
REQUEST_PROCESSING_ERROR = 'request_processing_error'
RESPONSE_NOT_ACCEPTED = "response_not_accepted"
RESPONSE_PROCESSING_ERROR = 'response_processing_error'


class Inviter(AbstractStateMachine):

    def __init__(self, *args, **kwargs):
        self.__problem_report = None
        super().__init__(*args, **kwargs)

    @property
    def protocols(self) -> List[str]:
        return [ConnProtocolMessage.PROTOCOL, Ack.PROTOCOL]

    @property
    def problem_report(self) -> ConnProblemReport:
        return self.__problem_report

    async def create_connection(
            self, me: Pairwise.Me, connection_key: str, request: ConnRequest, my_endpoint: Endpoint
    ) -> (bool, Pairwise):
        # Validate request
        request.validate()
        self.__problem_report = None
        # Step 1: Extract their info from connection request
        their_did, their_vk, their_endpoint_address, their_routing_keys = request.extract_their_info()
        invitee_endpoint = TheirEndpoint(
            endpoint=their_endpoint_address,
            verkey=their_vk,
            routing_keys=their_routing_keys
        )
        # Allocate transport channel between self and theirs by verkeys factor
        transport = await self.transports.spawn(me.verkey, invitee_endpoint)
        await transport.start(self.protocols, self.time_to_live)
        try:
            # Step 2: build connection response
            response = ConnResponse(did=me.did, verkey=me.verkey, endpoint=my_endpoint.address)
            if request.please_ack:
                response.thread_id = request.ack_message_id
            my_did_doc = response.did_doc
            await response.sign_connection(transport.wallet.crypto, connection_key)
            response.please_ack = True
            ok, ack = await transport.switch(response)
            if ok:
                if isinstance(ack, Ack) or isinstance(ack, Ping):
                    # Step 3: store their did
                    await transport.wallet.did.store_their_did(their_did, their_vk)
                    # Step 4: create pairwise
                    their = Pairwise.Their(
                        did=their_did,
                        label=request.label,
                        endpoint=their_endpoint_address,
                        verkey=their_vk,
                        routing_keys=their_routing_keys
                    )
                    metadata = {
                        'me': {
                            'did': me.did,
                            'verkey': me.verkey,
                            'did_doc': dict(my_did_doc)
                        },
                        'their': {
                            'did': their_did,
                            'verkey': their_vk,
                            'label': request.label,
                            'endpoint': {
                                'address': their_endpoint_address,
                                'routing_keys': their_routing_keys
                            },
                            'did_doc': dict(request.did_doc)
                        }
                    }
                    pairwise = Pairwise(me=me, their=their, metadata=metadata)
                    return True, pairwise
                elif isinstance(response, ConnProblemReport):
                    self.__problem_report = response
                    logging.error('Code: %s; Explain: %s' % (response.problem_code, response.explain))
                    return False, None
                else:
                    self.__problem_report = ConnProblemReport(
                        problem_code=REQUEST_PROCESSING_ERROR,
                        explain='Expect for connection response ack. Unexpected message type "%s"' % str(response.type),
                        thread_id=request.id
                    )
                    await transport.send(self.__problem_report)
                    return False, None
            else:
                self.__problem_report = ConnProblemReport(
                    problem_code=REQUEST_PROCESSING_ERROR,
                    explain='Response ack awaiting was terminated by timeout',
                    thread_id=request.id
                )
                await transport.send(self.__problem_report)
                return False, None
        finally:
            await transport.stop()


class Invitee(AbstractStateMachine):

    def __init__(self, *args, **kwargs):
        self.__problem_report = None
        super().__init__(*args, **kwargs)

    @property
    def protocols(self) -> List[str]:
        return [ConnProtocolMessage.PROTOCOL, Ack.PROTOCOL]

    @property
    def problem_report(self) -> ConnProblemReport:
        return self.__problem_report

    async def create_connection(
            self, me: Pairwise.Me, invitation: Invitation, my_label: str, my_endpoint: Endpoint
    ) -> (bool, Pairwise):
        # Validate invitation
        invitation.validate()
        self.__problem_report = None
        # Extract Inviter connection_key
        connection_key = invitation.recipient_keys[0]
        inviter_endpoint = TheirEndpoint(
            endpoint=invitation.endpoint,
            verkey=connection_key
        )
        # Allocate transport channel between self and theirs by verkeys factor
        transport = await self.transports.spawn(me.verkey, inviter_endpoint)
        await transport.start(self.protocols, self.time_to_live)
        try:
            # Step 1: send connection request to Inviter
            request = ConnRequest(
                label=my_label,
                did=me.did,
                verkey=me.verkey,
                endpoint=my_endpoint.address
            )
            request.please_ack = True
            ok, response = await transport.switch(request)
            if ok:
                if isinstance(response, ConnResponse):
                    # Step 2: process connection response from Inviter
                    success = await response.verify_connection(transport.wallet.crypto)
                    try:
                        response.validate()
                    except SiriusValidationError:
                        success = False
                    if success and (response['connection~sig']['signer'] == connection_key):
                        # Step 3: extract Inviter info and store did
                        their_did, their_vk, their_endpoint_address, their_routing_keys = response.extract_their_info()
                        await transport.wallet.did.store_their_did(their_did, their_vk)
                        # Step 4: Send ack to Inviter
                        if response.please_ack:
                            ack = Ack(thread_id=response.id, status=Status.OK)
                            await transport.send(ack)
                        else:
                            ping = Ping(comment='Connection established', response_requested=False)
                            await transport.send(ping)
                        # Step 5: Make Pairwise instance
                        their = Pairwise.Their(
                            did=their_did,
                            label=invitation.label,
                            endpoint=their_endpoint_address,
                            verkey=their_vk,
                            routing_keys=their_routing_keys
                        )
                        metadata = {
                            'me': {
                                'did': me.did,
                                'verkey': me.verkey,
                                'did_doc': dict(request.did_doc)
                            },
                            'their': {
                                'did': their_did,
                                'verkey': their_vk,
                                'label': invitation.label,
                                'endpoint': {
                                    'address': their_endpoint_address,
                                    'routing_keys': their_routing_keys
                                },
                                'did_doc': dict(response.did_doc)
                            }
                        }
                        pairwise = Pairwise(me=me, their=their, metadata=metadata)
                        return True, pairwise
                    else:
                        self.__problem_report = ConnProblemReport(
                            problem_code=RESPONSE_NOT_ACCEPTED,
                            explain='Invalid connection response signature for connection_key: "%s"' % connection_key,
                            thread_id=response.id
                        )
                        await transport.send(self.__problem_report)
                        return False, None
                elif isinstance(response, ConnProblemReport):
                    self.__problem_report = response
                    logging.error('Code: %s; Explain: %s' % (response.problem_code, response.explain))
                    return False, None
                else:
                    self.__problem_report = ConnProblemReport(
                        problem_code=RESPONSE_NOT_ACCEPTED,
                        explain='Expect for connection response. Unexpected message type "%s"' % str(response.type),
                        thread_id=response.id
                    )
                    await transport.send(self.__problem_report)
                    return False, None
            else:
                self.__problem_report = ConnProblemReport(
                    problem_code=RESPONSE_PROCESSING_ERROR,
                    explain='Response awaiting was terminated by timeout',
                )
                await transport.send(self.__problem_report)
                return False, None
        finally:
            await transport.stop()



import copy
from typing import List
from datetime import datetime

import pytest

from sirius_sdk import Agent
from sirius_sdk.agent.consensus.simple.state_machines import MicroLedgerSimpleConsensus
from sirius_sdk.agent.consensus.simple.messages import *

from .conftest import get_pairwise
from .helpers import run_coroutines


async def routine_of_ledger_creator(
        creator: Agent, me: Pairwise.Me, participants: List[str], ledger_name: str, genesis: List[dict]
):
    machine = MicroLedgerSimpleConsensus(
        creator.wallet.crypto, me, creator.pairwise_list, creator.microledgers, creator
    )
    genesis = [Transaction.create(txn) for txn in genesis]
    success, ledger = await machine.init_microledger(ledger_name, participants, genesis)
    return success, ledger


async def routine_of_ledger_creation_acceptor(acceptor: Agent):
    listener = await acceptor.subscribe()
    event = await listener.get_one()
    assert event.pairwise is not None
    propose = event.message
    assert isinstance(propose, InitRequestLedgerMessage)
    machine = MicroLedgerSimpleConsensus(
        acceptor.wallet.crypto, event.pairwise.me, acceptor.pairwise_list, acceptor.microledgers, acceptor
    )
    success, ledger = await machine.accept_microledger(event.pairwise, propose)
    return success, ledger


async def routine_of_txn_committer(
        creator: Agent, me: Pairwise.Me, participants: List[str], ledger: Microledger, txns: List[dict]
):
    machine = MicroLedgerSimpleConsensus(
        creator.wallet.crypto, me, creator.pairwise_list, creator.microledgers, creator
    )
    txns = [Transaction.create(txn) for txn in txns]
    success, txns = await machine.commit(ledger, participants, txns)
    return success, txns


async def routine_of_txn_acceptor(acceptor: Agent, txns: List[Transaction] = None):
    listener = await acceptor.subscribe()
    while True:
        event = await listener.get_one()
        assert event.pairwise is not None
        propose = event.message
        if isinstance(propose, ProposeTransactionsMessage):
            if txns:
                propose['transactions'] = txns
            machine = MicroLedgerSimpleConsensus(
                acceptor.wallet.crypto, event.pairwise.me, acceptor.pairwise_list, acceptor.microledgers, acceptor
            )
            success = await machine.accept_commit(event.pairwise, propose)
            return success


@pytest.mark.asyncio
async def test_init_ledger_messaging(A: Agent, B: Agent, ledger_name: str):
    await A.open()
    await B.open()
    try:
        A2B = await get_pairwise(A, B)
        B2A = await get_pairwise(B, A)
        A2B.me.did = 'did:peer:' + A2B.me.did
        B2A.me.did = 'did:peer:' + B2A.me.did
        genesis_txns = [
            Transaction({"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"})
        ]
        request = InitRequestLedgerMessage(
            participants=[A2B.me.did, B2A.me.did],
            ledger_name=ledger_name,
            genesis=genesis_txns,
            root_hash='xxx'
        )

        await request.add_signature(A.wallet.crypto, A2B.me)
        await request.add_signature(B.wallet.crypto, B2A.me)

        assert len(request.signatures) == 2

        await request.check_signatures(A.wallet.crypto, A2B.me.did)
        await request.check_signatures(A.wallet.crypto, B2A.me.did)
        await request.check_signatures(A.wallet.crypto)
        await request.check_signatures(B.wallet.crypto, A2B.me.did)
        await request.check_signatures(B.wallet.crypto, B2A.me.did)
        await request.check_signatures(B.wallet.crypto)

        response = InitResponseLedgerMessage()
        response.assign_from(request)

        payload1 = dict(**request)
        payload2 = dict(**response)
        assert payload1 != payload2

        del payload1['@id']
        del payload1['@type']
        del payload2['@id']
        del payload2['@type']
        assert payload1 == payload2
    finally:
        await A.close()
        await B.close()


@pytest.mark.asyncio
async def test_transaction_messaging(A: Agent, B: Agent, ledger_name: str):
    await A.open()
    await B.open()
    try:
        a2b = await get_pairwise(A, B)
        b2a = await get_pairwise(B, A)
        a2b.me.did = 'did:peer:' + a2b.me.did
        b2a.me.did = 'did:peer:' + b2a.me.did
        genesis_txns = [
            Transaction({"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"})
        ]
        ledger_for_a, txns = await A.microledgers.create(ledger_name, genesis_txns)
        ledger_for_b, txns = await B.microledgers.create(ledger_name, genesis_txns)

        new_transactions = [
            Transaction({"reqId": 2, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op2"}),
            Transaction({"reqId": 3, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3"}),
        ]
        pos1, pos2, new_txns = await ledger_for_a.append(new_transactions)
        # A -> B
        state = MicroLedgerState(
                {
                    'name': ledger_for_a.name,
                    'seq_no': ledger_for_a.seq_no,
                    'size': ledger_for_a.size,
                    'uncommitted_size': ledger_for_a.uncommitted_size,
                    'root_hash': ledger_for_a.root_hash,
                    'uncommitted_root_hash': ledger_for_a.uncommitted_root_hash
                }
        )
        x = MicroLedgerState.from_ledger(ledger_for_a)
        assert state == x
        assert state.hash == x.hash
        propose = ProposeTransactionsMessage(
            transactions=new_txns,
            state=state
        )
        propose.validate()
        # B -> A
        await ledger_for_b.append(propose.transactions)
        pre_commit = PreCommitTransactionsMessage(
            state=MicroLedgerState(
                {
                    'name': ledger_for_b.name,
                    'seq_no': ledger_for_b.seq_no,
                    'size': ledger_for_b.size,
                    'uncommitted_size': ledger_for_b.uncommitted_size,
                    'root_hash': ledger_for_b.root_hash,
                    'uncommitted_root_hash': ledger_for_b.uncommitted_root_hash
                }
            )
        )
        await pre_commit.sign_state(B.wallet.crypto, b2a.me)
        pre_commit.validate()
        ok, loaded_state_hash = await pre_commit.verify_state(A.wallet.crypto, a2b.their.verkey)
        assert ok is True
        assert loaded_state_hash == state.hash
        # A -> B
        commit = CommitTransactionsMessage()
        commit.add_pre_commit(a2b.their.did, pre_commit)
        commit.validate()
        states = await commit.verify_pre_commits(A.wallet.crypto, state)
        assert a2b.their.did in str(states)
        assert a2b.their.verkey in str(states)
        # B -> A (post-commit)
        post_commit = PostCommitTransactionsMessage()
        await post_commit.add_commit_sign(B.wallet.crypto, commit, b2a.me)
        post_commit.validate()
        ok = await post_commit.verify_commits(A.wallet.crypto, commit, [a2b.their.verkey])
        assert ok is True
    finally:
        await A.close()
        await B.close()


@pytest.mark.asyncio
async def test_simple_consensus_init_ledger(A: Agent, B: Agent, C: Agent, ledger_name: str):
    await A.open()
    await B.open()
    await C.open()
    try:
        A2B = await get_pairwise(A, B)
        A2C = await get_pairwise(A, C)
        assert A2B.me == A2C.me
        B2A = await get_pairwise(B, A)
        B2C = await get_pairwise(B, C)
        assert B2A.me == B2C.me
        C2A = await get_pairwise(C, A)
        C2B = await get_pairwise(C, B)
        assert C2A.me == C2B.me
        participants = [
            A2B.me.did,
            A2B.their.did,
            A2C.their.did
        ]
        genesis = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"}
        ]
        coro_creator = routine_of_ledger_creator(A, A2B.me, participants, ledger_name, genesis)
        coro_acceptor1 = routine_of_ledger_creation_acceptor(B)
        coro_acceptor2 = routine_of_ledger_creation_acceptor(C)

        stamp1 = datetime.now()
        print('> begin')
        await run_coroutines(coro_creator, coro_acceptor1, coro_acceptor2, timeout=30)
        print('> end')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'***** Consensus timeout: {delta.seconds}')

        is_exists_for_A = await A.microledgers.is_exists(ledger_name)
        is_exists_for_B = await B.microledgers.is_exists(ledger_name)
        is_exists_for_C = await C.microledgers.is_exists(ledger_name)
        assert is_exists_for_A
        assert is_exists_for_B
        assert is_exists_for_C

        for agent in [A, B, C]:
            ledger = await agent.microledgers.ledger(ledger_name)
            txns = await ledger.get_all_transactions()
            assert len(txns) == 2
            assert 'op1' in str(txns)
            assert 'op2' in str(txns)
    finally:
        await A.close()
        await B.close()
        await C.close()


@pytest.mark.asyncio
async def test_simple_consensus_commit(A: Agent, B: Agent, C: Agent, ledger_name: str):
    await A.open()
    await B.open()
    await C.open()
    try:
        A2B = await get_pairwise(A, B)
        A2C = await get_pairwise(A, C)
        assert A2B.me == A2C.me
        B2A = await get_pairwise(B, A)
        B2C = await get_pairwise(B, C)
        assert B2A.me == B2C.me
        C2A = await get_pairwise(C, A)
        C2B = await get_pairwise(C, B)
        assert C2A.me == C2B.me
        participants = [
            A2B.me.did,
            A2B.their.did,
            A2C.their.did
        ]
        genesis = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"}
        ]
        ledger_for_a, _ = await A.microledgers.create(ledger_name, genesis)
        await B.microledgers.create(ledger_name, genesis)
        await C.microledgers.create(ledger_name, genesis)

        txns = [
            {"reqId": 3, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3"},
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op5"},
        ]
        coro_committer = routine_of_txn_committer(A, A2B.me, participants, ledger_for_a, txns)
        coro_acceptor1 = routine_of_txn_acceptor(B)
        coro_acceptor2 = routine_of_txn_acceptor(C)

        stamp1 = datetime.now()
        print('> begin')
        await run_coroutines(coro_committer, coro_acceptor1, coro_acceptor2, timeout=60)
        print('> end')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'***** Consensus timeout: {delta.seconds}')

        ledger_for_a = await A.microledgers.ledger(ledger_name)
        ledger_for_b = await B.microledgers.ledger(ledger_name)
        ledger_for_c = await C.microledgers.ledger(ledger_name)
        for ledger in [ledger_for_a, ledger_for_b, ledger_for_c]:
            all_txns = await ledger.get_all_transactions()
            assert 'op3' in str(all_txns)
            assert 'op4' in str(all_txns)
            assert 'op5' in str(all_txns)
    finally:
        await A.close()
        await B.close()
        await C.close()


@pytest.mark.asyncio
async def test_simple_consensus_error(A: Agent, B: Agent, C: Agent, ledger_name: str):
    await A.open()
    await B.open()
    await C.open()
    try:
        A2B = await get_pairwise(A, B)
        A2C = await get_pairwise(A, C)
        assert A2B.me == A2C.me
        B2A = await get_pairwise(B, A)
        B2C = await get_pairwise(B, C)
        assert B2A.me == B2C.me
        C2A = await get_pairwise(C, A)
        C2B = await get_pairwise(C, B)
        assert C2A.me == C2B.me
        participants = [
            A2B.me.did,
            A2B.their.did,
            A2C.their.did
        ]
        genesis = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"}
        ]
        ledger_for_a, _ = await A.microledgers.create(ledger_name, genesis)
        initial_state_root_hash = ledger_for_a.root_hash
        await B.microledgers.create(ledger_name, genesis)
        await C.microledgers.create(ledger_name, genesis)

        txns = [
            {"reqId": 3, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3"},
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op5"},
        ]
        broken_txns = [
            {"reqId": 3, "identifier": "BROKEN-5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3", "txnMetadata": {"seqNo": 4}},
            {"reqId": 4, "identifier": "BROKEN-2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4", "txnMetadata": {"seqNo": 5}},
            {"reqId": 5, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op5", "txnMetadata": {"seqNo": 6}},
        ]
        broken_txns = [Transaction(txn) for txn in broken_txns]
        coro_committer = routine_of_txn_committer(A, A2B.me, participants, ledger_for_a, txns)
        coro_acceptor1 = routine_of_txn_acceptor(B, broken_txns)
        coro_acceptor2 = routine_of_txn_acceptor(C)

        stamp1 = datetime.now()
        print('> begin')
        results = await run_coroutines(coro_committer, coro_acceptor1, coro_acceptor2, timeout=60)
        print('> end')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'***** Consensus timeout: {delta.seconds}')
        for res in results:
            if type(res) is tuple:
                val = res[0]
            else:
                val = res
            assert val is False

        ledger_for_a = await A.microledgers.ledger(ledger_name)
        ledger_for_b = await B.microledgers.ledger(ledger_name)
        ledger_for_c = await C.microledgers.ledger(ledger_name)
        for ledger in [ledger_for_a, ledger_for_b, ledger_for_c]:
            await ledger.reload()
            assert ledger.root_hash == initial_state_root_hash
            all_txns = await ledger.get_all_transactions()
            assert 'op3' not in str(all_txns)
            assert 'op4' not in str(all_txns)
            assert 'op5' not in str(all_txns)
    finally:
        await A.close()
        await B.close()
        await C.close()


from typing import List, Union
from datetime import datetime, timedelta

from sirius_sdk.agent.pairwise import Pairwise
from sirius_sdk.agent.codec import encode
from sirius_sdk.agent.aries_rfc.utils import utc_to_str, str_to_utc
from sirius_sdk.agent.ledger import Schema, CredentialDefinition
from sirius_sdk.errors.indy_exceptions import WalletItemNotFound
from sirius_sdk.agent.wallet.abstract.anoncreds import AbstractAnonCreds
from sirius_sdk.agent.sm import AbstractStateMachine, StateMachineTerminatedWithError
from sirius_sdk.agent.aries_rfc.feature_0015_acks import Ack, Status
from sirius_sdk.agent.aries_rfc.feature_0036_issue_credential.messages import *


PROPOSE_NOT_ACCEPTED = "propose_not_accepted"
OFFER_PROCESSING_ERROR = 'offer_processing_error'
REQUEST_NOT_ACCEPTED = "request_not_accepted"
ISSUE_PROCESSING_ERROR = 'issue_processing_error'
RESPONSE_FOR_UNKNOWN_REQUEST = "response_for_unknown_request"


class Issuer(AbstractStateMachine):
    
    def __init__(
            self, api: AbstractAnonCreds, holder: Pairwise,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.__api = api
        self.__holder = holder
        self.__transport = None
        self.__problem_report = None

    async def issue(
            self, values: dict, schema: Schema, cred_def: CredentialDefinition,
            comment: str = None, locale: str = BaseIssueCredentialMessage.DEF_LOCALE,
            preview: List[Any] = None, translation: List[AttribTranslation] = None, cred_id: str = None
    ) -> bool:
        await self.__start()
        try:
            try:
                # Step-1: Send offer to holder
                offer = await self.__api.issuer_create_credential_offer(cred_def_id=cred_def.id)
                expires_time = datetime.utcnow() + timedelta(seconds=self.time_to_live)
                offer_msg = OfferCredentialMessage(
                    comment=comment,
                    locale=locale,
                    offer=offer,
                    cred_def=cred_def.body,
                    preview=preview,
                    issuer_schema=schema.body,
                    translation=translation,
                    expires_time=utc_to_str(expires_time)
                )
                offer_msg.please_ack = True
                resp = await self.__switch(offer_msg)
                if not isinstance(resp, RequestCredentialMessage):
                    raise StateMachineTerminatedWithError(OFFER_PROCESSING_ERROR, 'Unexpected @type: %s' % str(resp.type))
                # Step-2: Create credential
                request_msg = resp
                encoded_cred_values = dict()
                for key, value in values.items():
                    encoded_cred_values[key] = dict(raw=str(value), encoded=encode(value))
                ret = await self.__api.issuer_create_credential(
                    cred_offer=offer,
                    cred_req=request_msg.cred_request,
                    cred_values=encoded_cred_values,
                    rev_reg_id=None,
                    blob_storage_reader_handle=None
                )
                cred, cred_revoc_id, revoc_reg_delta = ret
                # Step-3: Issue and wait Ack
                issue_msg = IssueCredentialMessage(
                    comment=comment,
                    locale=locale,
                    cred=cred,
                    cred_id=cred_id
                )
                issue_msg.please_ack = True
                ack = await self.__switch(issue_msg)
                if not isinstance(ack, Ack):
                    raise StateMachineTerminatedWithError(ISSUE_PROCESSING_ERROR, 'Unexpected @type: %s' % str(resp.type))
            except StateMachineTerminatedWithError as e:
                self.__problem_report = IssueProblemReport(e.problem_code, e.explain)
                if e.notify:
                    await self.__send(self.__problem_report)
                return False
            else:
                return True
        finally:
            await self.__stop()

    @property
    def problem_report(self) -> IssueProblemReport:
        return self.__problem_report

    @property
    def protocols(self) -> List[str]:
        return [BaseIssueCredentialMessage.PROTOCOL, Ack.PROTOCOL]

    async def __start(self):
        self.__transport = await self.transports.spawn(self.__holder)
        await self.__transport.start(self.protocols, self.time_to_live)

    async def __stop(self):
        if self.__transport:
            await self.__transport.stop()
            self.__transport = None

    async def __switch(self, request: BaseIssueCredentialMessage) -> Union[BaseIssueCredentialMessage, Ack]:
        ok, resp = await self.__transport.switch(request)
        if ok:
            if isinstance(resp, BaseIssueCredentialMessage) or isinstance(resp, Ack):
                try:
                    resp.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(ISSUE_PROCESSING_ERROR, e.message)
                return resp
            elif isinstance(resp, IssueProblemReport):
                raise StateMachineTerminatedWithError(resp.problem_code, resp.explain, notify=False)
            else:
                raise StateMachineTerminatedWithError(ISSUE_PROCESSING_ERROR, 'Unexpected response @type: %s' % str(resp.type))
        else:
            raise StateMachineTerminatedWithError(ISSUE_PROCESSING_ERROR, 'Response awaiting terminated by timeout')

    async def __send(self, msg: Union[BaseIssueCredentialMessage, Ack, IssueProblemReport]):
        await self.__transport.send(msg)


class Holder(AbstractStateMachine):

    def __init__(
            self, api: AbstractAnonCreds, issuer: Pairwise,
            comment: str = None, locale: str = BaseIssueCredentialMessage.DEF_LOCALE,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.__api = api
        self.__issuer = issuer
        self.__problem_report = None
        self.__comment = comment
        self.__locale = locale

    async def accept(self, offer: OfferCredentialMessage, master_secret_id: str) -> (bool, Optional[str]):
        await self.__start()
        try:
            try:
                offer_msg = offer
                try:
                    offer_msg.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, e.message)
                # Step-1: Process Issuer Offer
                cred_request, cred_metadata = await self.__api.prover_create_credential_req(
                    prover_did=self.__issuer.me.did,
                    cred_offer=offer_msg.offer,
                    cred_def=offer_msg.cred_def,
                    master_secret_id=master_secret_id
                )
                # Step-2: Send request to Issuer
                request_msg = RequestCredentialMessage(
                    comment=self.__comment,
                    locale=self.__locale,
                    cred_request=cred_request
                )
                if offer_msg.please_ack:
                    request_msg.thread_id = offer_msg.id
                resp = await self.__switch(request_msg)
                if not isinstance(resp, IssueCredentialMessage):
                    raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, 'Unexpected @type: %s' % str(resp.type))
                issue_msg = resp
                try:
                    issue_msg.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, e.message)
                # Step-3: Store credential
                cred_id = await self._store_credential(
                    cred_metadata, issue_msg.cred, offer.cred_def, None, issue_msg.cred_id
                )
                ack = Ack(thread_id=issue_msg.id, status=Status.OK)
                await self.__send(ack)
            except StateMachineTerminatedWithError as e:
                self.__problem_report = IssueProblemReport(e.problem_code, e.explain)
                if e.notify:
                    await self.__send(self.__problem_report)
                return False, None
            else:
                return True, cred_id
        finally:
            await self.__stop()

    @property
    def problem_report(self) -> IssueProblemReport:
        return self.__problem_report

    @property
    def protocols(self) -> List[str]:
        return [BaseIssueCredentialMessage.PROTOCOL, Ack.PROTOCOL]

    async def _store_credential(
            self, cred_metadata: dict, cred: dict, cred_def: dict, rev_reg_def: Optional[dict], cred_id: Optional[str]
    ) -> str:
        try:
            cred_older = await self.__api.prover_get_credential(cred_id)
        except WalletItemNotFound:
            cred_older = None
        if cred_older:
            # Delete older credential
            await self.__api.prover_delete_credential(cred_id)
        cred_id = await self.__api.prover_store_credential(
            cred_req_metadata=cred_metadata,
            cred=cred,
            cred_def=cred_def,
            rev_reg_def=rev_reg_def,
            cred_id=cred_id
        )
        return cred_id

    async def __start(self):
        self.__transport = await self.transports.spawn(self.__issuer)
        await self.__transport.start(self.protocols, self.time_to_live)

    async def __stop(self):
        if self.__transport:
            await self.__transport.stop()
            self.__transport = None

    async def __switch(self, request: BaseIssueCredentialMessage) -> Union[BaseIssueCredentialMessage, Ack]:
        ok, resp = await self.__transport.switch(request)
        if ok:
            if isinstance(resp, BaseIssueCredentialMessage) or isinstance(resp, Ack):
                try:
                    resp.validate()
                except SiriusValidationError as e:
                    raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, e.message)
                return resp
            elif isinstance(resp, IssueProblemReport):
                raise StateMachineTerminatedWithError(resp.problem_code, resp.explain, notify=False)
            else:
                raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, 'Unexpected issuer request @type: %s' % str(resp.type))
        else:
            raise StateMachineTerminatedWithError(REQUEST_NOT_ACCEPTED, 'Issuer request awaiting terminated by timeout')

    async def __send(self, msg: Union[BaseIssueCredentialMessage, Ack, IssueProblemReport]):
        await self.__transport.send(msg)



import uuid
from datetime import datetime

import pytest

from sirius_sdk import Agent, Pairwise


@pytest.mark.asyncio
async def test_pairwise_list(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    try:
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        p = Pairwise(
            me=Pairwise.Me(
                did=did1, verkey=verkey1
            ),
            their=Pairwise.Their(
                did=did2, label='Test-Pairwise', endpoint='http://endpoint', verkey=verkey2
            ),
            metadata=dict(test='test-value')
        )

        lst1 = await agent1.wallet.pairwise.list_pairwise()
        await agent1.pairwise_list.ensure_exists(p)
        lst2 = await agent1.wallet.pairwise.list_pairwise()
        assert len(lst1) < len(lst2)

        ok = await agent1.pairwise_list.is_exists(did2)
        assert ok is True

        loaded = await agent1.pairwise_list.load_for_verkey(verkey2)
        assert loaded.metadata == p.metadata
    finally:
        await agent1.close()
        await agent2.close()



import uuid
from typing import List
from datetime import datetime

import pytest

from sirius_sdk import Agent, Pairwise
from sirius_sdk.agent.codec import encode
from sirius_sdk.agent.aries_rfc.utils import str_to_utc
from sirius_sdk.agent.ledger import CredentialDefinition
from sirius_sdk.agent.aries_rfc.feature_0037_present_proof import Verifier, Prover, AttribTranslation, \
    RequestPresentationMessage

from .conftest import get_pairwise
from .helpers import run_coroutines, IndyAgent


async def run_verifier(
        agent: Agent, prover: Pairwise, proof_request: dict, translation: List[AttribTranslation] = None
) -> bool:
    try:
        machine = Verifier(
            api=agent.wallet.anoncreds, cache=agent.wallet.cache, prover=prover, pool_name='default', transports=agent
        )
        success = await machine.verify(proof_request, translation=translation, comment='I am Verifier')
        return success
    except Exception as e:
        print('==== Verifier routine Exception: ' + repr(e))


async def run_prover(agent: Agent, verifier: Pairwise, master_secret_id: str):
    listener = await agent.subscribe()
    event = await listener.get_one()
    assert event.pairwise is not None
    assert event.pairwise.their.did == verifier.their.did
    request = event.message
    assert isinstance(request, RequestPresentationMessage)
    ttl = 60
    if request.expires_time:
        expire = str_to_utc(request.expires_time, raise_exceptions=False)
        delta = expire - datetime.utcnow()
        if delta.seconds > 0:
            ttl = delta.seconds
    try:
        machine = Prover(
            api=agent.wallet.anoncreds, cache=agent.wallet.cache, verifier=verifier,
            pool_name='default', transports=agent, time_to_live=ttl
        )
        success = await machine.prove(request, master_secret_id)
        return success
    except Exception as e:
        print('==== Prover routine Exception: ' + repr(e))


@pytest.mark.asyncio
async def test_sane(agent1: Agent, agent2: Agent, agent3: Agent):
    issuer = agent1
    prover = agent2
    verifier = agent3
    await issuer.open()
    await prover.open()
    await verifier.open()
    try:
        print('Establish pairwises')
        i2p = await get_pairwise(issuer, prover)
        p2i = await get_pairwise(prover, issuer)
        v2p = await get_pairwise(verifier, prover)
        p2v = await get_pairwise(prover, verifier)

        print('Register schema')
        did_issuer, verkey_issuer = i2p.me.did, i2p.me.verkey
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did_issuer, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = issuer.ledger('default')
        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did_issuer)
        assert ok is True

        print('Register credential def')
        ok, cred_def = await ledger.register_cred_def(
            cred_def=CredentialDefinition(tag='TAG', schema=schema),
            submitter_did=did_issuer
        )
        assert ok is True

        print('Prepare Prover')
        master_secret_name = 'secret-' + uuid.uuid4().hex
        prover_secret_id = await prover.wallet.anoncreds.prover_create_master_secret(master_secret_name)
        cred_values = {'attr1': 'Value-1', 'attr2': 456, 'attr3': 5.87}
        cred_id = 'cred-id-' + uuid.uuid4().hex

        # Issue credential
        offer = await issuer.wallet.anoncreds.issuer_create_credential_offer(cred_def_id=cred_def.id)
        cred_request, cred_metadata = await prover.wallet.anoncreds.prover_create_credential_req(
            prover_did=p2i.me.did, cred_offer=offer, cred_def=cred_def.body, master_secret_id=prover_secret_id
        )
        encoded_cred_values = dict()
        for key, value in cred_values.items():
            encoded_cred_values[key] = dict(raw=str(value), encoded=encode(value))
        ret = await issuer.wallet.anoncreds.issuer_create_credential(
            cred_offer=offer,
            cred_req=cred_request,
            cred_values=encoded_cred_values,
            rev_reg_id=None,
            blob_storage_reader_handle=None
        )
        cred, cred_revoc_id, revoc_reg_delta = ret
        await prover.wallet.anoncreds.prover_store_credential(
            cred_req_metadata=cred_metadata,
            cred=cred,
            cred_def=cred_def.body,
            rev_reg_def=None,
            cred_id=cred_id
        )

        # FIRE !!!
        attr_referent_id = 'attr1_referent'
        pred_referent_id = 'predicate1_referent'
        proof_request = {
            "nonce": await verifier.wallet.anoncreds.generate_nonce(),
            "name": "Test ProofRequest",
            "version": "0.1",
            "requested_attributes": {
                attr_referent_id: {
                    "name": "attr1",
                    "restrictions": {
                        "issuer_did": did_issuer
                    }
                }
            },
            "requested_predicates": {
                pred_referent_id: {
                    'name': 'attr2',
                    'p_type': '>=',
                    'p_value': 100,
                    "restrictions": {
                        "issuer_did": did_issuer
                    }
                }
            }
        }
        coro_verifier = run_verifier(
            agent=verifier,
            prover=v2p,
            proof_request=proof_request
        )
        coro_prover = run_prover(
            agent=prover,
            verifier=p2v,
            master_secret_id=prover_secret_id
        )
        print('Run state machines')
        results = await run_coroutines(coro_verifier, coro_prover, timeout=60)
        print('Finish state machines')
        print(str(results))
        assert len(results) == 2
        for res in results:
            assert res is True

    finally:
        await issuer.close()
        await prover.close()
        await verifier.close()


@pytest.mark.asyncio
async def test_back_compatibility(agent1: Agent, agent2: Agent, agent3: IndyAgent):
    issuer = agent1
    prover = agent3
    verifier = agent2
    await issuer.open()
    await verifier.open()
    try:
        pass
    finally:
        await issuer.close()
        await verifier.close()



import json
from urllib.parse import urlparse, urlunparse

import pytest

from sirius_sdk import Agent, Pairwise
from sirius_sdk.agent.connections import Endpoint
from sirius_sdk.agent.aries_rfc.feature_0160_connection_protocol import *

from .helpers import run_coroutines, IndyAgent


def replace_url_components(url: str, base: str = None) -> str:
    ret = url
    if base:
        parsed = urlparse(url)
        components = list(parsed)
        components[1] = urlparse(base).netloc
        ret = urlunparse(components)
    return ret


async def read_events(agent: Agent):
    listener = await agent.subscribe()
    async for event in listener:
        print('========= EVENT ============')
        print(json.dumps(event, indent=2, sort_keys=True))
        print('============================')


async def run_inviter(agent: Agent, expected_connection_key: str, me: Pairwise.Me=None, replace_endpoints: bool=False):
    my_endpoint = [e for e in agent.endpoints if e.routing_keys == []][0]
    if replace_endpoints:
        new_address = replace_url_components(my_endpoint.address, pytest.test_suite_overlay_address)
        my_endpoint = Endpoint(new_address, my_endpoint.routing_keys, is_default=my_endpoint.is_default)
    listener = await agent.subscribe()
    async for event in listener:
        connection_key = event['recipient_verkey']
        if expected_connection_key == connection_key:
            request = event['message']
            assert isinstance(request, ConnRequest)
            if replace_endpoints:
                request['connection']['did_doc']['service'][0]['serviceEndpoint'] = replace_url_components(
                    request['connection']['did_doc']['service'][0]['serviceEndpoint'],
                    pytest.old_agent_overlay_address
                )
            # Setup state machine
            machine = Inviter(agent)
            if me is None:
                my_did, my_verkey = await agent.wallet.did.create_and_store_my_did()
                me = Pairwise.Me(did=my_did, verkey=my_verkey)
            # create connection
            ok, pairwise = await machine.create_connection(me, connection_key, request, my_endpoint)
            assert ok is True
            await agent.pairwise_list.ensure_exists(pairwise)
    pass


async def run_invitee(agent: Agent, invitation: Invitation, my_label: str, me: Pairwise.Me=None, replace_endpoints: bool=False):
    if me is None:
        my_did, my_verkey = await agent.wallet.did.create_and_store_my_did()
        me = Pairwise.Me(did=my_did, verkey=my_verkey)
    my_endpoint = [e for e in agent.endpoints if e.routing_keys == []][0]
    if replace_endpoints:
        new_address = replace_url_components(my_endpoint.address, pytest.test_suite_overlay_address)
        my_endpoint = Endpoint(new_address, my_endpoint.routing_keys, is_default=my_endpoint.is_default)
        new_address = replace_url_components(invitation['serviceEndpoint'], pytest.old_agent_overlay_address)
        invitation['serviceEndpoint'] = new_address
    # Create and start machine
    machine = Invitee(agent)
    ok, pairwise = await machine.create_connection(
        me=me, invitation=invitation, my_label=my_label, my_endpoint=my_endpoint
    )
    assert ok is True
    await agent.pairwise_list.ensure_exists(pairwise)


@pytest.mark.asyncio
async def test_establish_connection(agent1: Agent, agent2: Agent):
    inviter = agent1
    invitee = agent2
    await inviter.open()
    await invitee.open()
    try:
        # Get endpoints
        inviter_endpoint_address = [e for e in inviter.endpoints if e.routing_keys == []][0].address
        connection_key = await inviter.wallet.crypto.create_key()
        invitation = Invitation(label='Inviter', endpoint=inviter_endpoint_address, recipient_keys=[connection_key])
        # Init Me
        did, verkey = await inviter.wallet.did.create_and_store_my_did()
        inviter_me = Pairwise.Me(did, verkey)
        did, verkey = await invitee.wallet.did.create_and_store_my_did()
        invitee_me = Pairwise.Me(did, verkey)

        await run_coroutines(
            run_inviter(inviter, connection_key, inviter_me),
            run_invitee(invitee, invitation, 'Invitee', invitee_me)
        )

        # Check for Inviter
        pairwise = await inviter.wallet.pairwise.get_pairwise(invitee_me.did)
        assert pairwise['my_did'] == inviter_me.did
        pairwise = await inviter.pairwise_list.load_for_verkey(invitee_me.verkey)
        assert pairwise is not None
        assert pairwise.their.did == invitee_me.did
        # Check for Invitee
        pairwise = await invitee.wallet.pairwise.get_pairwise(inviter_me.did)
        assert pairwise['my_did'] == invitee_me.did
        pairwise = await invitee.pairwise_list.load_for_verkey(inviter_me.verkey)
        assert pairwise is not None
        assert pairwise.their.did == inviter_me.did

    finally:
        await inviter.close()
        await invitee.close()


@pytest.mark.asyncio
async def test_update_pairwise_metadata(agent1: Agent, agent2: Agent):
    inviter = agent1
    invitee = agent2
    await inviter.open()
    await invitee.open()
    try:
        # Init Me
        did, verkey = await inviter.wallet.did.create_and_store_my_did()
        inviter_side = Pairwise.Me(did, verkey)
        did, verkey = await invitee.wallet.did.create_and_store_my_did()
        invitee_side = Pairwise.Me(did, verkey)
        # Manually set pairwise list
        await inviter.wallet.did.store_their_did(invitee_side.did, invitee_side.verkey)
        await invitee.wallet.did.store_their_did(inviter_side.did, inviter_side.verkey)
        await inviter.wallet.pairwise.create_pairwise(
            their_did=invitee_side.did,
            my_did=inviter_side.did
        )
        await invitee.wallet.pairwise.create_pairwise(
            their_did=inviter_side.did,
            my_did=invitee_side.did
        )

        # Run
        inviter_endpoint_address = [e for e in inviter.endpoints if e.routing_keys == []][0].address
        connection_key = await inviter.wallet.crypto.create_key()
        invitation = Invitation(label='Inviter', endpoint=inviter_endpoint_address, recipient_keys=[connection_key])

        await run_coroutines(
            run_inviter(inviter, connection_key, inviter_side),
            run_invitee(invitee, invitation, 'Invitee', invitee_side)
        )

        # Check for Inviter
        pairwise = await inviter.wallet.pairwise.get_pairwise(invitee_side.did)
        assert pairwise['metadata'] != {}

        # Check for Invitee
        pairwise = await invitee.wallet.pairwise.get_pairwise(inviter_side.did)
        assert pairwise['metadata'] != {}

    finally:
        await inviter.close()
        await invitee.close()


@pytest.mark.asyncio
async def test_invitee_back_compatibility(indy_agent: IndyAgent, agent1: Agent):
    their_invitaton = await indy_agent.create_invitation(label='Test Invitee')
    invitation = Invitation.from_url(their_invitaton['url'])
    inviter = agent1
    await inviter.open()
    try:
        my_did, my_verkey = await agent1.wallet.did.create_and_store_my_did()
        me = Pairwise.Me(did=my_did, verkey=my_verkey)
        await run_coroutines(
            run_invitee(agent1, invitation, 'Invitee', me, True),
            read_events(agent1)
        )
        invitation_pairwise = None
        for pairwise in await agent1.wallet.pairwise.list_pairwise():
            if pairwise['my_did'] == my_did:
                invitation_pairwise = pairwise
                break
        assert invitation_pairwise is not None
    finally:
        await inviter.close()


@pytest.mark.asyncio
async def test_inviter_back_compatibility(indy_agent: IndyAgent, agent1: Agent):
    await agent1.open()
    try:
        # Get endpoints
        my_endpoint_address = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        connection_key = await agent1.wallet.crypto.create_key()
        my_endpoint_address = replace_url_components(my_endpoint_address, pytest.test_suite_overlay_address)
        my_invitation = Invitation(label='Inviter', endpoint=my_endpoint_address, recipient_keys=[connection_key])
        url = my_invitation.invitation_url
        my_did, my_verkey = await agent1.wallet.did.create_and_store_my_did()
        me = Pairwise.Me(did=my_did, verkey=my_verkey)
        await run_coroutines(
            run_inviter(agent1, connection_key, me, True),
            indy_agent.invite(invitation_url=url),
        )
        invitated_pairwise = None
        for pairwise in await agent1.wallet.pairwise.list_pairwise():
            if pairwise['my_did'] == my_did:
                invitated_pairwise = pairwise
                break
        assert invitated_pairwise is not None
    finally:
        await agent1.close()



from sirius_sdk.agent.connections import AgentRPC
from sirius_sdk.agent.wallet.impl.did import DIDProxy
from sirius_sdk.agent.wallet.impl.cache import CacheProxy
from sirius_sdk.agent.wallet.impl.crypto import CryptoProxy
from sirius_sdk.agent.wallet.impl.ledger import LedgerProxy
from sirius_sdk.agent.wallet.impl.pairwise import PairwiseProxy
from sirius_sdk.agent.wallet.impl.anoncreds import AnonCredsProxy
from sirius_sdk.agent.wallet.impl.non_secrets import NonSecretsProxy


class DynamicWallet:

    def __init__(self, rpc: AgentRPC):
        self.__rpc = rpc
        self.__did = DIDProxy(rpc)
        self.__crypto = CryptoProxy(rpc)
        self.__cache = CacheProxy(rpc)
        self.__pairwise = PairwiseProxy(rpc)
        self.__non_secrets = NonSecretsProxy(rpc)
        self.__ledger = LedgerProxy(rpc)
        self.__anoncreds = AnonCredsProxy(rpc)

    @property
    def did(self) -> DIDProxy:
        return self.__did

    @property
    def crypto(self) -> CryptoProxy:
        return self.__crypto

    @property
    def cache(self) -> CacheProxy:
        return self.__cache

    @property
    def ledger(self) -> LedgerProxy:
        return self.__ledger

    @property
    def pairwise(self) -> PairwiseProxy:
        return self.__pairwise

    @property
    def anoncreds(self) -> AnonCredsProxy:
        return self.__anoncreds

    @property
    def non_secrets(self) -> NonSecretsProxy:
        return self.__non_secrets

    async def generate_wallet_key(self, seed: str = None) -> str:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/generate_wallet_key',
            params=dict(seed=seed)
        )



import json

import pytest

from sirius_sdk.encryption import create_keypair, pack_message, unpack_message, bytes_to_b58, P2PConnection


@pytest.mark.asyncio
def test_sane():
    verkey, sigkey = create_keypair(b'000000000000000000000000000SEED1')
    verkey_recipient = bytes_to_b58(verkey)
    sigkey_recipient = bytes_to_b58(sigkey)
    verkey, sigkey = create_keypair(b'000000000000000000000000000SEED2')
    verkey_sender = bytes_to_b58(verkey)
    sigkey_sender = bytes_to_b58(sigkey)

    message = {
        'content': 'Test encryption строка'
    }
    message = json.dumps(message)

    packed = pack_message(
        message=message,
        to_verkeys=[verkey_recipient],
        from_verkey=verkey_sender,
        from_sigkey=sigkey_sender
    )
    unpacked, sender_vk, recip_vk = unpack_message(
        enc_message=packed,
        my_verkey=verkey_recipient,
        my_sigkey=sigkey_recipient
    )
    assert message == unpacked
    assert sender_vk, verkey_sender
    assert recip_vk, verkey_recipient


@pytest.mark.asyncio
def test_fixture():
    verkey, sigkey = create_keypair(b'000000000000000000000000000SEED1')
    verkey_recipient = bytes_to_b58(verkey)
    sigkey_recipient = bytes_to_b58(sigkey)
    verkey, sigkey = create_keypair(b'000000000000000000000000000SEED2')
    verkey_sender = bytes_to_b58(verkey)
    packed = b'{"protected": "eyJlbmMiOiAieGNoYWNoYTIwcG9seTEzMDVfaWV0ZiIsICJ0eXAiOiAiSldNLzEuMCIsICJhbGciOiAiQXV0aGNyeXB0IiwgInJlY2lwaWVudHMiOiBbeyJlbmNyeXB0ZWRfa2V5IjogInBKcW1xQS1IVWR6WTNWcFFTb2dySGx4WTgyRnc3Tl84YTFCSmtHU2VMT014VUlwT0RQWTZsMVVsaVVvOXFwS0giLCAiaGVhZGVyIjogeyJraWQiOiAiM1ZxZ2ZUcDZRNFZlRjhLWTdlVHVXRFZBWmFmRDJrVmNpb0R2NzZLR0xtZ0QiLCAic2VuZGVyIjogIjRlYzhBeFRHcWtxamd5NHlVdDF2a0poeWlYZlNUUHo1bTRKQjk1cGZSMG1JVW9KajAwWmswNmUyUEVDdUxJYmRDck8xeTM5LUhGTG5NdW5YQVJZWk5rZ2pyYV8wYTBQODJpbVdNcWNHc1FqaFd0QUhOcUw1OGNkUUYwYz0iLCAiaXYiOiAiVU1PM2o1ZHZwQnFMb2Rvd3V0c244WEMzTkVqSWJLb2oifX1dfQ==", "iv": "MchkHF2M-4hneeUJ", "ciphertext": "UgcdsV-0rIkP25eJuRSROOuqiTEXp4NToKjPMmqqtJs-Ih1b5t3EEbrrHxeSfPsHtlO6J4OqA1jc5uuD3aNssUyLug==", "tag": "sQD8qgJoTrRoyQKPeCSBlQ=="}'
    unpacked, sender_vk, recip_vk = unpack_message(
        enc_message=packed,
        my_verkey=verkey_recipient,
        my_sigkey=sigkey_recipient
    )
    message = json.dumps({
        'content': 'Test encryption строка'
    })
    assert message == unpacked
    assert sender_vk, verkey_sender
    assert recip_vk, verkey_recipient



import asyncio
from abc import ABC, abstractmethod
from typing import List, Union, Optional

from multipledispatch import dispatch

from sirius_sdk.messaging import Message
from sirius_sdk.encryption import P2PConnection
from sirius_sdk.storages import AbstractImmutableCollection
from sirius_sdk.agent.listener import Listener
from sirius_sdk.agent.pairwise import Pairwise, TheirEndpoint
from sirius_sdk.agent.wallet.wallets import DynamicWallet
from sirius_sdk.agent.ledger import Ledger
from sirius_sdk.agent.pairwise import AbstractPairwiseList, WalletPairwiseList
from sirius_sdk.agent.storages import InWalletImmutableCollection
from sirius_sdk.agent.microledgers import MicroledgerList
from sirius_sdk.agent.coprotocols import PairwiseCoProtocolTransport, ThreadBasedCoProtocolTransport, TheirEndpointCoProtocolTransport
from sirius_sdk.agent.connections import AgentRPC, AgentEvents, BaseAgentConnection, Endpoint


class TransportLayers(ABC):

    @dispatch(str, TheirEndpoint)
    @abstractmethod
    async def spawn(self, my_verkey: str, endpoint: TheirEndpoint) -> TheirEndpointCoProtocolTransport:
        raise NotImplemented

    @dispatch(Pairwise)
    @abstractmethod
    async def spawn(self, pairwise: Pairwise) -> PairwiseCoProtocolTransport:
        raise NotImplemented

    @dispatch(str, Pairwise)
    @abstractmethod
    async def spawn(self, thid: str, pairwise: Pairwise) -> ThreadBasedCoProtocolTransport:
        raise NotImplemented

    @dispatch(str)
    @abstractmethod
    async def spawn(self, thid: str) -> ThreadBasedCoProtocolTransport:
        raise NotImplemented

    @dispatch(str, Pairwise, str)
    @abstractmethod
    async def spawn(self, thid: str, pairwise: Pairwise, pthid: str) -> ThreadBasedCoProtocolTransport:
        raise NotImplemented

    @dispatch(str, str)
    @abstractmethod
    async def spawn(self, thid: str, pthid: str) -> ThreadBasedCoProtocolTransport:
        raise NotImplemented


class Agent(TransportLayers):
    """
    Agent connection in the self-sovereign identity ecosystem.

    Managing an identity is complex. It is implementation of tools to help you to develop SSI Smart-Contracts logic.
    See details:
      - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0004-agents
    """

    def __init__(
            self, server_address: str, credentials: bytes,
            p2p: P2PConnection, timeout: int=BaseAgentConnection.IO_TIMEOUT, loop: asyncio.AbstractEventLoop=None,
            storage: AbstractImmutableCollection=None, name: str=None
    ):
        """
        :param server_address: example https://my-cloud-provider.com
        :param credentials: credentials that point websocket connection to your agent and server-side services like
          routing keys maintenance ant etc.
        :param p2p: encrypted connection to establish tunnel to Agent that is running on server-side
        """
        self.__server_address = server_address
        self.__credentials = credentials
        self.__p2p = p2p
        self.__rpc = None
        self.__events = None
        self.__wallet = None
        self.__timeout = timeout
        self.__loop = loop
        self.__endpoints = []
        self.__ledgers = {}
        self.__storage = storage
        self.__pairwise_list = None
        self.__microledgers = None
        self.__name = name

    @property
    def name(self) -> Optional[str]:
        return self.__name

    @property
    def wallet(self) -> DynamicWallet:
        """Indy wallet keys/schemas/CredDefs maintenance"""
        self.__check_is_open()
        return self.__wallet

    def ledger(self, name: str) -> Ledger:
        self.__check_is_open()
        return self.__ledgers.get(name, None)

    @property
    def endpoints(self) -> List[Endpoint]:
        self.__check_is_open()
        return self.__endpoints

    @property
    def microledgers(self) -> MicroledgerList:
        self.__check_is_open()
        return self.__microledgers

    @property
    def pairwise_list(self) -> AbstractPairwiseList:
        self.__check_is_open()
        return self.__pairwise_list

    @dispatch(str, TheirEndpoint)
    async def spawn(self, my_verkey: str, endpoint: TheirEndpoint) -> TheirEndpointCoProtocolTransport:
        new_rpc = await AgentRPC.create(
            self.__server_address, self.__credentials, self.__p2p, self.__timeout, self.__loop
        )
        return TheirEndpointCoProtocolTransport(
            my_verkey=my_verkey,
            endpoint=endpoint,
            rpc=new_rpc
        )

    @dispatch(Pairwise)
    async def spawn(self, pairwise: Pairwise) -> PairwiseCoProtocolTransport:
        new_rpc = await AgentRPC.create(
            self.__server_address, self.__credentials, self.__p2p, self.__timeout, self.__loop
        )
        return PairwiseCoProtocolTransport(
            pairwise=pairwise,
            rpc=new_rpc
        )

    @dispatch(str, Pairwise)
    async def spawn(self, thid: str, pairwise: Pairwise) -> ThreadBasedCoProtocolTransport:
        new_rpc = await AgentRPC.create(
            self.__server_address, self.__credentials, self.__p2p, self.__timeout, self.__loop
        )
        return ThreadBasedCoProtocolTransport(
            thid=thid,
            pairwise=pairwise,
            rpc=new_rpc
        )

    @dispatch(str)
    @abstractmethod
    async def spawn(self, thid: str) -> ThreadBasedCoProtocolTransport:
        new_rpc = await AgentRPC.create(
            self.__server_address, self.__credentials, self.__p2p, self.__timeout, self.__loop
        )
        return ThreadBasedCoProtocolTransport(
            thid=thid,
            pairwise=None,
            rpc=new_rpc
        )

    @dispatch(str, Pairwise, str)
    async def spawn(self, thid: str, pairwise: Pairwise, pthid: str) -> ThreadBasedCoProtocolTransport:
        new_rpc = await AgentRPC.create(
            self.__server_address, self.__credentials, self.__p2p, self.__timeout, self.__loop
        )
        return ThreadBasedCoProtocolTransport(
            thid=thid,
            pairwise=pairwise,
            rpc=new_rpc,
            pthid=pthid
        )

    @dispatch(str, str)
    @abstractmethod
    async def spawn(self, thid: str, pthid: str) -> ThreadBasedCoProtocolTransport:
        new_rpc = await AgentRPC.create(
            self.__server_address, self.__credentials, self.__p2p, self.__timeout, self.__loop
        )
        return ThreadBasedCoProtocolTransport(
            thid=thid,
            pairwise=None,
            rpc=new_rpc,
            pthid=pthid
        )

    async def open(self):
        self.__rpc = await AgentRPC.create(
            self.__server_address, self.__credentials, self.__p2p, self.__timeout, self.__loop
        )
        self.__endpoints = self.__rpc.endpoints
        self.__wallet = DynamicWallet(rpc=self.__rpc)
        if self.__storage is None:
            self.__storage = InWalletImmutableCollection(self.__wallet.non_secrets)
        for network in self.__rpc.networks:
            self.__ledgers[network] = Ledger(
                name=network, api=self.__wallet.ledger,
                issuer=self.__wallet.anoncreds, cache=self.__wallet.cache, storage=self.__storage
            )
        self.__pairwise_list = WalletPairwiseList(api=(self.__wallet.pairwise, self.__wallet.did))
        self.__microledgers = MicroledgerList(api=self.__rpc)

    async def subscribe(self) -> Listener:
        self.__check_is_open()
        self.__events = await AgentEvents.create(
            self.__server_address, self.__credentials, self.__p2p, self.__timeout, self.__loop
        )
        return Listener(self.__events, self.pairwise_list)

    async def close(self):
        if self.__rpc:
            await self.__rpc.close()
        if self.__events:
            await self.__events.close()
        self.__wallet = None

    async def ping(self) -> bool:
        success = await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/ping_agent'
        )
        return success

    async def send_message(
            self, message: Message, their_vk: Union[List[str], str],
            endpoint: str, my_vk: Optional[str], routing_keys: Optional[List[str]] = None
    ) -> (bool, Message):
        """
        Implementation of basicmessage feature
        See details:
          - https://github.com/hyperledger/aries-rfcs/tree/master/features/0095-basic-message

        :param message: Message
          See details:
           - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0020-message-types
        :param their_vk: Verkey of recipient
        :param endpoint: Endpoint address of recipient
        :param my_vk: VerKey of Sender (AuthCrypt mode)
          See details:
           - https://github.com/hyperledger/aries-rfcs/tree/master/features/0019-encryption-envelope#authcrypt-mode-vs-anoncrypt-mode
        :param routing_keys: Routing key of recipient
        """
        self.__check_is_open()
        await self.__rpc.send_message(
            message=message, their_vk=their_vk, endpoint=endpoint,
            my_vk=my_vk, routing_keys=routing_keys, coprotocol=False
        )

    async def send_to(self, message: Message, to: Pairwise):
        """Implementation of basicmessage feature
        See details:
          - https://github.com/hyperledger/aries-rfcs/tree/master/features/0095-basic-message

        :param message: Message
          See details:
           - https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0020-message-types
        :param to: Pairwise (P2P) connection that have been established outside
        """
        self.__check_is_open()
        await self.send_message(
            message=message,
            their_vk=to.their.verkey,
            endpoint=to.their.endpoint,
            my_vk=to.me.verkey,
            routing_keys=to.their.routing_keys
        )

    async def generate_qr_code(self, value: str) -> str:
        self.__check_is_open()
        resp = await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/admin/1.0/generate_qr',
            params={
                'value': value
            }
        )
        return resp['url']

    def __check_is_open(self):
        if self.__rpc and self.__rpc.is_open:
            return self.__endpoints
        else:
            raise RuntimeError('Open Agent at first!')



import uuid

import pytest

from sirius_sdk import Agent
from sirius_sdk.agent.storages import InWalletImmutableCollection
from sirius_sdk.storages import InMemoryKeyValueStorage, InMemoryImmutableCollection


@pytest.mark.asyncio
async def test_inmemory_kv_storage():
    kv = InMemoryKeyValueStorage()
    await kv.select_db('db1')

    await kv.set('key1', 'value1')
    value = await kv.get('key1')
    assert value == 'value1'

    await kv.select_db('db2')
    await kv.set('key1', 1000)
    value = await kv.get('key1')
    assert value == 1000

    await kv.select_db('db1')
    value = await kv.get('key1')
    assert value == 'value1'

    await kv.delete('key1')
    value = await kv.get('key1')
    assert value is None

    await kv.delete('unknown-key')


@pytest.mark.asyncio
async def test_inmemory_immutable_collection():
    collection = InMemoryImmutableCollection()

    await collection.select_db('db1')
    await collection.add('Value1', {'tag1': 'tag-val-1', 'tag2': 'tag-val-2'})
    await collection.add('Value2', {'tag1': 'tag-val-1', 'tag2': 'tag-val-3'})

    fetched1 = await collection.fetch({'tag1': 'tag-val-1'})
    assert len(fetched1) == 2

    fetched1 = await collection.fetch({'tag2': 'tag-val-2'})
    assert len(fetched1) == 1
    assert fetched1[0] == 'Value1'

    await collection.select_db('db2')
    fetched3 = await collection.fetch({})
    assert len(fetched3) == 0


@pytest.mark.asyncio
async def test_inwallet_immutable_collection(agent1: Agent):
    await agent1.open()
    try:
        collection = InWalletImmutableCollection(agent1.wallet.non_secrets)

        value1 = {
            'key1': 'value1',
            'key2': 10000
        }
        value2 = {
            'key1': 'value2',
            'key2': 50000
        }

        await collection.select_db(db_name=uuid.uuid4().hex)
        await collection.add(value1, {'tag': 'value1'})
        await collection.add(value2, {'tag': 'value2'})

        fetched, count = await collection.fetch({'tag': 'value1'})
        assert count == 1
        assert len(fetched) == 1
        assert fetched[0] == value1

        fetched, count = await collection.fetch({'tag': 'value2'})
        assert count == 1
        assert len(fetched) == 1
        assert fetched[0] == value2

        fetched, count = await collection.fetch({})
        assert count == 2

        await collection.select_db(db_name=uuid.uuid4().hex)
        fetched, count = await collection.fetch({})
        assert count == 0
    finally:
        await agent1.close()



import json
import uuid
import base64
from typing import List, Optional
from collections import UserDict

from sirius_sdk.errors.exceptions import *
from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage, AriesProblemReport, THREAD_DECORATOR


CREDENTIAL_TRANSLATION_TYPE = "https://github.com/Sirius-social/agent/tree/master/messages/credential-translation"
CREDENTIAL_TRANSLATION_ID = "credential-translation"


class ProposedAttrib(UserDict):

    def __init__(
            self, name: str, value: str = None, mime_type: str = None, referent: str = None, cred_def_id: str = None
    ):
        super().__init__()
        self.data['name'] = name
        if mime_type:
            self.data['mime-type'] = mime_type
        if value:
            self.data['value'] = value
        if referent:
            self.data['referent'] = referent
        if cred_def_id:
            self.data['cred_def_id'] = cred_def_id

    def to_json(self):
        return self.data


class ProposedPredicate(UserDict):

    def __init__(self, name: str, predicate: str, threshold, cred_def_id: str = None):
        super().__init__()
        self.data['name'] = name
        self.data['predicate'] = predicate
        self.data['threshold'] = threshold
        if cred_def_id:
            self.data['cred_def_id'] = cred_def_id

    def to_json(self):
        return self.data


class AttribTranslation(UserDict):

    def __init__(self, attrib_name: str, translation: str, **kwargs):
        super().__init__()
        self.data['attrib_name'] = attrib_name
        self.data['translation'] = translation

    def to_json(self):
        return self.data


class BasePresentProofMessage(AriesProtocolMessage, metaclass=RegisterMessage):

    PROTOCOL = 'present-proof'
    DEF_LOCALE = 'en'

    def __init__(self, locale: str = DEF_LOCALE, *args, **kwargs):
        version = kwargs.pop('version', '1.1')
        super().__init__(version=version, *args, **kwargs)
        self['~l10n'] = {"locale": locale}

    @property
    def locale(self) -> str:
        return self.get('~l10n', {}).get('locale', self.DEF_LOCALE)

    @property
    def ack_message_id(self) -> str:
        return self.get('~please_ack', {}).get('message_id', None) or self.id

    @property
    def please_ack(self) -> bool:
        """https://github.com/hyperledger/aries-rfcs/tree/master/features/0317-please-ack"""
        return self.get('~please_ack', None) is not None

    @please_ack.setter
    def please_ack(self, flag: bool):
        if flag:
            self['~please_ack'] = {'message_id': self.id}
        elif '~please_ack' in self:
            del self['~please_ack']

    @property
    def thread_id(self) -> Optional[str]:
        return self.get('THREAD_DECORATOR', {}).get('thid', None)

    @thread_id.setter
    def thread_id(self, thid: str):
        thread = self.get(THREAD_DECORATOR, {})
        thread['thid'] = thid
        self[THREAD_DECORATOR] = thread


class PresentProofProblemReport(AriesProblemReport, metaclass=RegisterMessage):

    PROTOCOL = BasePresentProofMessage.PROTOCOL


class RequestPresentationMessage(BasePresentProofMessage):

    NAME = 'request-presentation'

    def __init__(
            self, proof_request: dict = None, comment: str = None,
            translation: List[AttribTranslation] = None, expires_time: str = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if comment:
            self['comment'] = comment
        id_suffix = uuid.uuid4().hex
        if proof_request:
            self["request_presentations~attach"] = [
                {
                    "@id": "libindy-request-presentation-" + id_suffix,
                    "mime-type": "application/json",
                    "data": {
                        "base64": base64.b64encode(json.dumps(proof_request).encode()).decode()
                    }
                }
            ]
        if translation:
            translation = [AttribTranslation(**item) for item in translation]
            self['~attach'] = [
                {
                    "@type": CREDENTIAL_TRANSLATION_TYPE,
                    "id": CREDENTIAL_TRANSLATION_ID,
                    '~l10n': {"locale": self.locale},
                    "mime-type": "application/json",
                    "data": {
                        "json": [trans.to_json() for trans in translation]
                    }
                }
            ]
        if expires_time:
            self['~timing'] = {
                "expires_time": expires_time
            }

    @property
    def proof_request(self) -> Optional[dict]:
        attaches = self.get('request_presentations~attach', None)
        if not attaches:
            return None
        if type(attaches) is dict:
            attaches = [attaches]
        accum = {}
        for attach in attaches:
            payload = json.loads(base64.b64decode(attach['data']['base64']).decode())
            accum.update(payload)
        return accum

    @property
    def comment(self) -> Optional[str]:
        return self.get('comment', None)

    @property
    def translation(self) -> Optional[List[AttribTranslation]]:
        attaches = self.get('~attach', [])
        tr = None
        for item in attaches:
            if item.get('@type', None) == CREDENTIAL_TRANSLATION_TYPE:
                tr = item
                break
        if tr:
            translation = tr.get('data', {}).get('json', [])
            return [AttribTranslation(**item) for item in translation]
        else:
            return None

    @property
    def expires_time(self) -> Optional[str]:
        return self.get('~timing', {}).get('expires_time', None)


class PresentationMessage(BasePresentProofMessage):

    NAME = 'presentation'

    def __init__(self, proof: dict = None, comment: str = None, presentation_id: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if comment:
            self['comment'] = comment
        if proof:
            if not presentation_id:
                presentation_id = uuid.uuid4().hex
            self["presentations~attach"] = [
                {
                    "@id": "libindy-presentation-" + presentation_id,
                    "mime-type": "application/json",
                    "data": {
                        "base64": base64.b64encode(json.dumps(proof).encode()).decode()
                    }
                }
            ]

    @property
    def proof(self) -> Optional[dict]:
        attaches = self.get('presentations~attach', None)
        if not attaches:
            return None
        if type(attaches) is dict:
            attaches = [attaches]
        accum = {}
        for attach in attaches:
            payload = json.loads(base64.b64decode(attach['data']['base64']).decode())
            accum.update(payload)
        return accum

    @property
    def comment(self) -> Optional[str]:
        return self.get('comment', None)



from typing import Optional, Any

from sirius_sdk.storages.abstract.kv_storage import AbstractKeyValueStorage


class InMemoryKeyValueStorage(AbstractKeyValueStorage):

    def __init__(self, *args, **kwargs):
        self.__databases = {}
        self.__selected_db = None
        super().__init__(*args, **kwargs)

    async def select_db(self, db_name: str):
        if db_name not in self.__databases:
            self.__databases[db_name] = {}
        self.__selected_db = self.__databases[db_name]

    async def set(self, key: str, value: Any):
        self.__selected_db[key] = value

    async def get(self, key: str) -> Optional[Any]:
        return self.__selected_db.get(key, None)

    async def delete(self, key: str):
        if key in self.__selected_db:
            del self.__selected_db[key]



from abc import ABC, abstractmethod
from inspect import iscoroutinefunction
from typing import List

from sirius_sdk.errors.exceptions import BaseSiriusException
from sirius_sdk.agent.agent import TransportLayers


class AbstractStateMachine(ABC):

    def __init__(self, transports: TransportLayers, time_to_live: int = 60, logger=None):
        """
        :param transports: aries-rfc transports factory
        :param time_to_live: state machine time to live to finish progress
        """
        self.__transports = transports
        self.__time_to_live = time_to_live
        if logger is not None:
            if iscoroutinefunction(logger) or callable(logger):
                pass
            else:
                raise RuntimeError('Expect logger is iscoroutine function or callable object')
        self.__logger = logger

    @property
    def transports(self) -> TransportLayers:
        return self.__transports

    @property
    def time_to_live(self) -> int:
        return self.__time_to_live

    @property
    @abstractmethod
    def protocols(self) -> List[str]:
        raise NotImplemented('Need to be implemented in descendant')

    async def log(self, **kwargs) -> bool:
        if self.__logger:
            kwargs = dict(**kwargs)
            kwargs['state_machine_id'] = id(self)
            await self.__logger(**kwargs)
        else:
            return False


class StateMachineTerminatedWithError(BaseSiriusException):

    def __init__(self, problem_code: str, explain: str, notify: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.problem_code = problem_code
        self.explain = explain
        self.notify = notify



import uuid
from typing import List, Any
from datetime import datetime, timedelta

import pytest

from sirius_sdk import Agent, Pairwise
from sirius_sdk.agent.wallet import NYMRole
from sirius_sdk.agent.aries_rfc.utils import str_to_utc
from sirius_sdk.agent.ledger import Schema, CredentialDefinition
from sirius_sdk.agent.aries_rfc.feature_0036_issue_credential import Issuer, Holder, AttribTranslation, \
    ProposedAttrib, OfferCredentialMessage

from .conftest import get_pairwise
from .helpers import run_coroutines, IndyAgent


async def run_issuer(
        agent: Agent, holder: Pairwise, values: dict, schema: Schema, cred_def: CredentialDefinition,
        preview: List[ProposedAttrib] = None, translation: List[AttribTranslation] = None, cred_id: str = None
):
    machine = Issuer(
        api=agent.wallet.anoncreds, holder=holder, transports=agent
    )
    success = await machine.issue(
        values=values,
        schema=schema,
        cred_def=cred_def,
        comment='Hello Iam issuer',
        preview=preview,
        translation=translation,
        cred_id=cred_id
    )
    return success


async def run_holder(agent: Agent, issuer: Pairwise, master_secret_id: str):
    listener = await agent.subscribe()
    event = await listener.get_one()
    assert event.pairwise is not None
    offer = event.message
    assert isinstance(offer, OfferCredentialMessage)
    ttl = 60
    if offer.expires_time:
        expire = str_to_utc(offer.expires_time, raise_exceptions=False)
        delta = expire - datetime.utcnow()
        if delta.seconds > 0:
            ttl = delta.seconds
    machine = Holder(
        api=agent.wallet.anoncreds, issuer=issuer, comment='Hello, Iam holder', transports=agent, time_to_live=ttl
    )
    success, cred_id = await machine.accept(offer, master_secret_id)
    return success, cred_id


async def run_issuer_indy_agent(
        indy_agent: IndyAgent, cred_def_id: str, cred_def: dict, values: dict, their_did: str,
        issuer_schema: dict = None, preview: List[ProposedAttrib] = None, translation: List[AttribTranslation] = None,
        rev_reg_id: str = None, cred_id: str = None, ttl: int = 60
) -> bool:
    log = await indy_agent.issue_credential(
        cred_def_id, cred_def, values, their_did, 'Test issuer', None, issuer_schema, preview, translation,
        rev_reg_id, cred_id, ttl
    )
    if len(log) > 2:
        last = log[-1]['message']
        pred = log[-2]['message']
        return ('Received ACK' in pred) and ('Done' in last)
    else:
        return False


@pytest.mark.asyncio
async def test_sane(agent1: Agent, agent2: Agent):
    issuer = agent1
    holder = agent2
    await issuer.open()
    await holder.open()
    try:
        i2h = await get_pairwise(issuer, holder)
        h2i = await get_pairwise(holder, issuer)

        did_issuer, verkey_issuer = i2h.me.did, i2h.me.verkey
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did_issuer, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = issuer.ledger('default')
        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did_issuer)
        assert ok is True

        ok, cred_def = await ledger.register_cred_def(
            cred_def=CredentialDefinition(tag='TAG', schema=schema),
            submitter_did=did_issuer
        )
        assert ok is True
        master_secret_name = 'secret-' + uuid.uuid4().hex
        holder_secret_id = await holder.wallet.anoncreds.prover_create_master_secret(master_secret_name)

        cred_id = 'cred-id-' + uuid.uuid4().hex
        coro_issuer = run_issuer(
            agent=issuer, holder=i2h,
            values={'attr1': 'Value-1', 'attr2': 567, 'attr3': 5.7},
            schema=schema, cred_def=cred_def, cred_id=cred_id
        )
        coro_holder = run_holder(
            agent=holder,
            issuer=h2i,
            master_secret_id=holder_secret_id
        )

        results = await run_coroutines(coro_issuer, coro_holder, timeout=60)
        cred_id = None
        for res in results:
            if type(res) is tuple:
                ok, cred_id = res
            else:
                ok = res
            assert ok is True

        assert cred_id is not None
        cred = await holder.wallet.anoncreds.prover_get_credential(cred_id)
        assert cred

    finally:
        await issuer.close()
        await holder.close()


@pytest.mark.asyncio
async def test_issuer_back_compatibility(indy_agent: IndyAgent, agent1: Agent):
    issuer = agent1
    await issuer.open()
    try:
        endpoint_issuer = [e for e in issuer.endpoints if e.routing_keys == []][0].address
        did_issuer, verkey_issuer = await issuer.wallet.did.create_and_store_my_did()
        did_holder, verkey_holder = await indy_agent.create_and_store_my_did()
        pairwise_for_issuer = Pairwise(
            me=Pairwise.Me(did_issuer, verkey_issuer),
            their=Pairwise.Their(did_holder, 'Holder', indy_agent.endpoint['url'], verkey_holder)
        )
        pairwise_for_holder = Pairwise(
            me=Pairwise.Me(did_holder, verkey_holder),
            their=Pairwise.Their(did_issuer, 'Issuer', endpoint_issuer, verkey_issuer)
        )
        pairwise_for_issuer.their.netloc = pytest.old_agent_overlay_address.replace('http://', '')
        pairwise_for_holder.their.netloc = pytest.test_suite_overlay_address.replace('http://', '')
        await indy_agent.create_pairwise_statically(pairwise_for_holder)
        await issuer.wallet.did.store_their_did(did_holder, verkey_holder)
        await issuer.pairwise_list.ensure_exists(pairwise_for_issuer)

        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did_issuer, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = issuer.ledger('default')
        ok, resp = await issuer.wallet.ledger.write_nym(
            'default', 'Th7MpTaRZVRYnPiabds81Y', did_issuer, verkey_issuer, 'Issuer', NYMRole.TRUST_ANCHOR
        )
        assert ok is True
        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did_issuer)
        assert ok is True

        ok, cred_def = await ledger.register_cred_def(
            cred_def=CredentialDefinition(tag='TAG', schema=schema),
            submitter_did=did_issuer
        )
        assert ok is True

        cred_id = 'cred-id-' + uuid.uuid4().hex
        coro_issuer = run_issuer(
            agent=issuer, holder=pairwise_for_issuer,
            values={'attr1': 'Value-1', 'attr2': 567, 'attr3': 5.7},
            schema=schema, cred_def=cred_def, cred_id=cred_id
        )

        results = await run_coroutines(coro_issuer, timeout=60)
        assert len(results) == 1
        assert results[0] is True
    finally:
        await issuer.close()


@pytest.mark.asyncio
async def test_holder_back_compatibility(indy_agent: IndyAgent, agent1: Agent):
    holder = agent1
    await holder.open()
    try:
        endpoint_holder = [e for e in holder.endpoints if e.routing_keys == []][0].address
        did_issuer, verkey_issuer = await indy_agent.create_and_store_my_did()
        did_holder, verkey_holder = await holder.wallet.did.create_and_store_my_did()
        pairwise_for_issuer = Pairwise(
            me=Pairwise.Me(did_issuer, verkey_issuer),
            their=Pairwise.Their(did_holder, 'Holder', endpoint_holder, verkey_holder)
        )
        pairwise_for_holder = Pairwise(
            me=Pairwise.Me(did_holder, verkey_holder),
            their=Pairwise.Their(did_issuer, 'Issuer', indy_agent.endpoint['url'], verkey_issuer)
        )
        pairwise_for_issuer.their.netloc = pytest.test_suite_overlay_address.replace('http://', '')
        pairwise_for_holder.their.netloc = pytest.old_agent_overlay_address.replace('http://', '')
        await indy_agent.create_pairwise_statically(pairwise_for_issuer)
        await holder.wallet.did.store_their_did(did_issuer, verkey_issuer)
        await holder.pairwise_list.ensure_exists(pairwise_for_holder)

        ok, resp = await agent1.wallet.ledger.write_nym(
            'default', 'Th7MpTaRZVRYnPiabds81Y', did_issuer, verkey_issuer, 'Issuer', NYMRole.TRUST_ANCHOR
        )
        assert ok is True
        # Register schema
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, schema = await indy_agent.register_schema(did_issuer, schema_name, '1.0', ['attr1', 'attr2', 'attr3'])
        cred_def_id, cred_def = await indy_agent.register_cred_def(did_issuer, schema_id, 'TAG')

        master_secret_name = 'secret-' + uuid.uuid4().hex
        holder_secret_id = await holder.wallet.anoncreds.prover_create_master_secret(master_secret_name)
        cred_id = 'cred-id-' + uuid.uuid4().hex

        coro_issuer = run_issuer_indy_agent(
            indy_agent=indy_agent, cred_def_id=cred_def_id, cred_def=cred_def,
            values={'attr1': 'Value-1', 'attr2': 567, 'attr3': 5.7},
            their_did=pairwise_for_issuer.their.did,
            issuer_schema=schema, rev_reg_id=None,
            cred_id=cred_id
        )
        coro_holder = run_holder(
            agent=holder,
            issuer=pairwise_for_holder,
            master_secret_id=holder_secret_id
        )

        results = await run_coroutines(coro_issuer, coro_holder, timeout=60)
        assert len(results) == 2

        for res in results:
            if type(res) is tuple:
                ok, cred_id = res
            else:
                ok = res
            assert ok is True

        assert cred_id is not None
        cred = await holder.wallet.anoncreds.prover_get_credential(cred_id)
        assert cred

    finally:
        await holder.close()



from typing import Optional

from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage, THREAD_DECORATOR


class Ping(AriesProtocolMessage, metaclass=RegisterMessage):
    """Implementation of Ping part for trust_ping protocol

    https://github.com/hyperledger/aries-rfcs/tree/master/features/0048-trust-ping
    """

    PROTOCOL = 'trust_ping'
    NAME = 'ping'

    def __init__(self, comment: Optional[str]=None, response_requested: Optional[bool]=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if comment is not None:
            self['comment'] = comment
        if response_requested is not None:
            self['response_requested'] = response_requested

    @property
    def comment(self) -> Optional[str]:
        return self.get('comment', None)

    @property
    def response_requested(self) -> Optional[bool]:
        return self.get('response_requested', None)

    @response_requested.setter
    def response_requested(self, value: bool):
        if value is True:
            self['response_requested'] = True
        else:
            self['response_requested'] = False


class Pong(AriesProtocolMessage, metaclass=RegisterMessage):
    """Implementation of Pong part for trust_ping protocol

    https://github.com/hyperledger/aries-rfcs/tree/master/features/0048-trust-ping
    """

    PROTOCOL = 'trust_ping'
    NAME = 'ping_response'

    def __init__(self, ping_id: str=None, comment: Optional[str]=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if ping_id is not None:
            self.get(THREAD_DECORATOR, {}).update({'thid': ping_id})
        if comment is not None:
            self['comment'] = comment

    @property
    def comment(self) -> Optional[str]:
        return self.get('comment', None)

    @property
    def ping_id(self):
        return self.get(THREAD_DECORATOR, {}).get('thid', None)



import json
from typing import Tuple, Union

from sirius_sdk.errors.exceptions import SiriusCryptoError
from sirius_sdk.encryption import pack_message, unpack_message


class P2PConnection:
    """"
    Pairwise static connection compatible with Indy SDK
    """

    def __init__(self, my_keys: Tuple[str, str], their_verkey: str):
        """
        :param my_keys: (verkey, sigkey) for encrypt/decrypt operations
        :param their_verkey: verkey of the counterparty
        """
        self.__my_keys = my_keys
        self.__their_verkey = their_verkey

    @property
    def my_verkey(self):
        return self.__my_keys[0]

    @property
    def their_verkey(self):
        return self.__their_verkey

    def pack(self, message: dict) -> bytes:
        """
        Encrypt message

        :param message:
        :return: encrypted message
        """
        packed = pack_message(
            message=json.dumps(message),
            to_verkeys=[self.__their_verkey],
            from_verkey=self.__my_keys[0],
            from_sigkey=self.__my_keys[1]
        )
        return packed

    def unpack(self, enc_message: Union[bytes, dict]) -> dict:
        """
        Decrypt message

        :param enc_message: encoded message
        :return: decrypted message
        """
        try:
            message, sender_vk, recip_vk = unpack_message(
                enc_message=enc_message,
                my_verkey=self.__my_keys[0],
                my_sigkey=self.__my_keys[1]
            )
        except ValueError as e:
            raise SiriusCryptoError(str(e))
        except KeyError as e:
            raise SiriusCryptoError(str(e))
        else:
            return json.loads(message)



import uuid
import asyncio

import pytest

from sirius_sdk import Agent
from sirius_sdk.agent.coprotocols import *
from .helpers import run_coroutines
from .helpers import ServerTestSuite


TEST_MSG_TYPES = [
    'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test_protocol/1.0/request-1',
    'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test_protocol/1.0/response-1',
    'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test_protocol/1.0/request-2',
    'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test_protocol/1.0/response-2',
]
MSG_LOG = []


async def routine1(protocol: AbstractCoProtocolTransport):
    first_req = Message({
        '@type': TEST_MSG_TYPES[0],
        'content': 'Request1'
    })
    MSG_LOG.append(first_req)
    ok, resp1 = await protocol.switch(message=first_req)
    assert ok is True
    MSG_LOG.append(resp1)
    ok, resp2 = await protocol.switch(
        message=Message({
            '@type': TEST_MSG_TYPES[2],
            'content': 'Request2'
        })
    )
    assert ok is True
    MSG_LOG.append(resp2)


async def routine2(protocol: AbstractCoProtocolTransport):
    await asyncio.sleep(1)
    ok, resp1 = await protocol.switch(
        message=Message({
            '@type': TEST_MSG_TYPES[1],
            'content': 'Response1'
        })
    )
    assert ok is True
    MSG_LOG.append(resp1)
    await protocol.send(
        message=Message({
            '@type': TEST_MSG_TYPES[3],
            'content': 'End'
        })
    )


def check_msg_log():
    assert len(MSG_LOG) == len(TEST_MSG_TYPES)
    for i, item in enumerate(TEST_MSG_TYPES):
        assert MSG_LOG[i].type == TEST_MSG_TYPES[i]
    assert MSG_LOG[0]['content'] == 'Request1'
    assert MSG_LOG[1]['content'] == 'Response1'
    assert MSG_LOG[2]['content'] == 'Request2'
    assert MSG_LOG[3]['content'] == 'End'


@pytest.mark.asyncio
async def test__their_endpoint_protocol(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Make protocol instances
        their1 = TheirEndpoint(agent2_endpoint, entity2['verkey'])
        agent1_protocol = await agent1.spawn(entity1['verkey'], their1)
        assert isinstance(agent1_protocol, TheirEndpointCoProtocolTransport)
        their2 = TheirEndpoint(agent1_endpoint, entity1['verkey'])
        agent2_protocol = await agent2.spawn(entity2['verkey'], their2)
        assert isinstance(agent2_protocol, TheirEndpointCoProtocolTransport)
        await agent1_protocol.start(['test_protocol'])
        await agent2_protocol.start(['test_protocol'])
        try:
            MSG_LOG.clear()
            await run_coroutines(routine1(agent1_protocol), routine2(agent2_protocol))
            check_msg_log()
        finally:
            await agent1_protocol.stop()
            await agent2_protocol.stop()
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test__pairwise_protocol(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        await agent1.wallet.did.store_their_did(did2, verkey2)
        await agent1.wallet.pairwise.create_pairwise(
            their_did=did2, my_did=did1
        )
        await agent2.wallet.did.store_their_did(did1, verkey1)
        await agent2.wallet.pairwise.create_pairwise(
            their_did=did1, my_did=did2
        )
        # Init pairwise list #2
        pairwise1 = Pairwise(
            me=Pairwise.Me(
                did=did1,
                verkey=verkey1
            ),
            their=Pairwise.Their(
                did=did2,
                label='Label-2',
                endpoint=agent2_endpoint,
                verkey=verkey2
            )
        )
        pairwise2 = Pairwise(
            me=Pairwise.Me(
                did=did2,
                verkey=verkey2
            ),
            their=Pairwise.Their(
                did=did1,
                label='Label-1',
                endpoint=agent1_endpoint,
                verkey=verkey1
            )
        )

        agent1_protocol = await agent1.spawn(pairwise1)
        agent2_protocol = await agent2.spawn(pairwise2)
        assert isinstance(agent1_protocol, PairwiseCoProtocolTransport)
        assert isinstance(agent2_protocol, PairwiseCoProtocolTransport)

        await agent1_protocol.start(['test_protocol'])
        await agent2_protocol.start(['test_protocol'])
        try:
            MSG_LOG.clear()
            await run_coroutines(routine1(agent1_protocol), routine2(agent2_protocol))
            check_msg_log()
        finally:
            await agent1_protocol.stop()
            await agent2_protocol.stop()
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test__threadbased_protocol(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        await agent1.wallet.did.store_their_did(did2, verkey2)
        await agent1.wallet.pairwise.create_pairwise(
            their_did=did2, my_did=did1
        )
        await agent2.wallet.did.store_their_did(did1, verkey1)
        await agent2.wallet.pairwise.create_pairwise(
            their_did=did1, my_did=did2
        )
        # Init pairwise list #2
        pairwise1 = Pairwise(
            me=Pairwise.Me(
                did=did1,
                verkey=verkey1
            ),
            their=Pairwise.Their(
                did=did2,
                label='Label-2',
                endpoint=agent2_endpoint,
                verkey=verkey2
            )
        )
        pairwise2 = Pairwise(
            me=Pairwise.Me(
                did=did2,
                verkey=verkey2
            ),
            their=Pairwise.Their(
                did=did1,
                label='Label-1',
                endpoint=agent1_endpoint,
                verkey=verkey1
            )
        )

        thread_id = uuid.uuid4().hex
        agent1_protocol = await agent1.spawn(thread_id, pairwise1)
        agent2_protocol = await agent2.spawn(thread_id, pairwise2)
        assert isinstance(agent1_protocol, ThreadBasedCoProtocolTransport)
        assert isinstance(agent2_protocol, ThreadBasedCoProtocolTransport)

        await agent1_protocol.start()
        await agent2_protocol.start()
        try:
            MSG_LOG.clear()
            await run_coroutines(routine1(agent1_protocol), routine2(agent2_protocol))
            check_msg_log()
        finally:
            await agent1_protocol.stop()
            await agent2_protocol.stop()
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test__protocols_intersections(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=15,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=15,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        await agent1.wallet.did.store_their_did(did2, verkey2)
        await agent1.wallet.pairwise.create_pairwise(
            their_did=did2, my_did=did1
        )
        await agent2.wallet.did.store_their_did(did1, verkey1)
        await agent2.wallet.pairwise.create_pairwise(
            their_did=did1, my_did=did2
        )
        # Init pairwise list #2
        pairwise1 = Pairwise(
            me=Pairwise.Me(
                did=did1,
                verkey=verkey1
            ),
            their=Pairwise.Their(
                did=did2,
                label='Label-2',
                endpoint=agent2_endpoint,
                verkey=verkey2
            )
        )
        pairwise2 = Pairwise(
            me=Pairwise.Me(
                did=did2,
                verkey=verkey2
            ),
            their=Pairwise.Their(
                did=did1,
                label='Label-1',
                endpoint=agent1_endpoint,
                verkey=verkey1
            )
        )

        thread_id = uuid.uuid4().hex
        agent1_protocol_threaded = await agent1.spawn(thread_id, pairwise1)
        agent2_protocol_threaded = await agent2.spawn(thread_id, pairwise2)
        assert isinstance(agent1_protocol_threaded, ThreadBasedCoProtocolTransport)
        assert isinstance(agent2_protocol_threaded, ThreadBasedCoProtocolTransport)
        agent1_protocol_pairwise = await agent1.spawn(pairwise1)
        agent2_protocol_pairwise = await agent2.spawn(pairwise2)
        assert isinstance(agent1_protocol_pairwise, PairwiseCoProtocolTransport)
        assert isinstance(agent2_protocol_pairwise, PairwiseCoProtocolTransport)

        await agent1_protocol_threaded.start(['test_protocol'])
        await agent2_protocol_threaded.start(['test_protocol'])
        await agent1_protocol_pairwise.start(['test_protocol'])
        await agent2_protocol_pairwise.start(['test_protocol'])
        try:
            MSG_LOG.clear()
            await run_coroutines(
                routine1(agent1_protocol_threaded), routine2(agent2_protocol_threaded),
                routine1(agent1_protocol_pairwise), routine2(agent2_protocol_pairwise),
            )
            # collect messages
            threaded_sequence = [msg for msg in MSG_LOG if '~thread' in msg]
            non_threaded_sequence = [msg for msg in MSG_LOG if '~thread' not in msg]
            # threaded messages
            MSG_LOG.clear()
            MSG_LOG.extend(threaded_sequence)
            check_msg_log()
            for msg in threaded_sequence:
                assert msg['~thread']['thid'] == thread_id
            # non-threaded messages
            MSG_LOG.clear()
            MSG_LOG.extend(non_threaded_sequence)
            check_msg_log()
        finally:
            await agent1_protocol_threaded.stop()
            await agent2_protocol_threaded.stop()
            await agent1_protocol_pairwise.stop()
            await agent2_protocol_pairwise.stop()
    finally:
        await agent1.close()
        await agent2.close()



import uuid
import json

import pytest

from sirius_sdk import Agent
from sirius_sdk.agent.wallet import RetrieveRecordOptions, CacheOptions, NYMRole


@pytest.mark.asyncio
async def test_crypto_pack_message(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    wallet_sender = agent1.wallet
    wallet_recipient = agent2.wallet
    try:
        verkey_sender = await wallet_sender.crypto.create_key()
        verkey_recipient = await wallet_recipient.crypto.create_key()
        assert verkey_recipient
        assert verkey_sender
        message = dict(content='Hello!')
        # 1: anon crypt mode
        wired_message1 = await wallet_sender.crypto.pack_message(message, [verkey_recipient])
        unpacked_message1 = await wallet_recipient.crypto.unpack_message(wired_message1)
        assert json.dumps(unpacked_message1['message'], sort_keys=True) == json.dumps(message, sort_keys=True)
        # 2: auth crypt mode
        wired_message2 = await wallet_sender.crypto.pack_message(message, [verkey_recipient], verkey_sender)
        unpacked_message2 = await wallet_recipient.crypto.unpack_message(wired_message2)
        assert json.dumps(unpacked_message2['message'], sort_keys=True) == json.dumps(message, sort_keys=True)
        assert wired_message2.decode() != wired_message1.decode()
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_crypto_sign(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    wallet_signer = agent1.wallet
    wallet_verifier = agent2.wallet
    try:
        key_signer = await wallet_signer.crypto.create_key()
        message = dict(content='Hello!')
        message_bytes = json.dumps(message).encode()
        signature = await wallet_signer.crypto.crypto_sign(key_signer, message_bytes)
        is_ok = await wallet_verifier.crypto.crypto_verify(key_signer, message_bytes, signature)
        assert is_ok is True

        key_signer2 = await wallet_signer.crypto.create_key()
        broken_signature = await wallet_signer.crypto.crypto_sign(key_signer, message_bytes)
        is_ok = await wallet_verifier.crypto.crypto_verify(key_signer2, message_bytes, broken_signature)
        assert is_ok is False
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_did_maintenance(agent1: Agent):
    await agent1.open()
    try:
        # 1: Create Key
        random_key = await agent1.wallet.did.create_key()
        assert random_key

        # 2: Set metadata
        metadata = dict(
            key1='value1',
            key2='value2'
        )
        await agent1.wallet.did.set_key_metadata(random_key, metadata)
        expected = json.dumps(metadata, sort_keys=True)
        metadata = await agent1.wallet.did.get_key_metadata(random_key)
        actual = json.dumps(metadata, sort_keys=True)
        assert expected == actual

        # 3: Create DID + Verkey
        did, verkey = await agent1.wallet.did.create_and_store_my_did()
        fully = await agent1.wallet.did.qualify_did(did, 'peer')
        assert did in fully

        # 4: Replace verkey
        verkey_new = await agent1.wallet.did.replace_keys_start(did=fully)
        assert verkey_new
        metadata_list = await agent1.wallet.did.list_my_dids_with_meta()
        assert any([m['tempVerkey'] == verkey_new for m in metadata_list])
        await agent1.wallet.did.replace_keys_apply(did=fully)
        metadata_list = await agent1.wallet.did.list_my_dids_with_meta()
        assert any([m['verkey'] == verkey_new for m in metadata_list])

        actual_verkey = await agent1.wallet.did.key_for_local_did(did=fully)
        assert verkey_new == actual_verkey
    finally:
        await agent1.close()


@pytest.mark.asyncio
async def test_their_did_maintenance(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    wallet_me = agent1.wallet
    wallet_their = agent2.wallet
    try:
        did_my, verkey_my = await wallet_me.did.create_and_store_my_did()
        did_their, verkey_their = await wallet_their.did.create_and_store_my_did()
        await wallet_me.did.store_their_did(did_their, verkey_their)
        metadata = dict(
            key1='value1',
            key2='value2'
        )
        await wallet_me.did.set_did_metadata(did_their, metadata)
        expected = json.dumps(metadata, sort_keys=True)
        metadata = await wallet_me.did.get_did_metadata(did_their)
        actual = json.dumps(metadata, sort_keys=True)
        assert expected == actual

        verkey = await wallet_me.did.key_for_local_did(did_their)
        assert verkey_their == verkey

        verkey_their_new = await wallet_their.did.replace_keys_start(did=did_their)
        await wallet_their.did.replace_keys_apply(did=did_their)
        await wallet_me.did.store_their_did(did_their, verkey_their_new)
        verkey = await wallet_me.did.key_for_local_did(did_their)
        assert verkey_their_new == verkey
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_record_value(agent1: Agent):
    await agent1.open()
    try:
        value = 'my-value-' + uuid.uuid4().hex
        my_id = 'my-id-' + uuid.uuid4().hex
        await agent1.wallet.non_secrets.add_wallet_record('type', my_id, value)
        opts = RetrieveRecordOptions()
        opts.check_all()
        value_info = await agent1.wallet.non_secrets.get_wallet_record('type', my_id, opts)
        assert value_info['id'] == my_id
        assert value_info['tags'] == {}
        assert value_info['value'] == value
        assert value_info['type'] == 'type'

        value_new = 'my-new-value-' + uuid.uuid4().hex
        await agent1.wallet.non_secrets.update_wallet_record_value('type', my_id, value_new)
        value_info = await agent1.wallet.non_secrets.get_wallet_record('type', my_id, opts)
        assert value_info['value'] == value_new

        await agent1.wallet.non_secrets.delete_wallet_record('type', my_id)
    finally:
        await agent1.close()


@pytest.mark.asyncio
async def test_record_value_with_tags(agent1: Agent):
    await agent1.open()
    try:
        value = 'my-value-' + uuid.uuid4().hex
        my_id = 'my-id-' + uuid.uuid4().hex
        tags = {
            'tag1': 'val1',
            '~tag2': 'val2'
        }
        await agent1.wallet.non_secrets.add_wallet_record('type', my_id, value, tags)
        opts = RetrieveRecordOptions()
        opts.check_all()
        value_info = await agent1.wallet.non_secrets.get_wallet_record('type', my_id, opts)
        assert value_info['id'] == my_id
        assert value_info['tags'] == tags
        assert value_info['value'] == value
        assert value_info['type'] == 'type'

        upd_tags = {
            'ext-tag': 'val3'
        }
        await agent1.wallet.non_secrets.update_wallet_record_tags('type', my_id, upd_tags)
        value_info = await agent1.wallet.non_secrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == upd_tags

        await agent1.wallet.non_secrets.add_wallet_record_tags('type', my_id, tags)
        value_info = await agent1.wallet.non_secrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == dict(**upd_tags, **tags)

        await agent1.wallet.non_secrets.delete_wallet_record_tags('type', my_id, ['ext-tag'])
        value_info = await agent1.wallet.non_secrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == tags
    finally:
        await agent1.close()


@pytest.mark.asyncio
async def test_record_value_with_tags_then_update(agent1: Agent):
    await agent1.open()
    try:
        value = 'my-value-' + uuid.uuid4().hex
        my_id = 'my-id-' + uuid.uuid4().hex
        await agent1.wallet.non_secrets.add_wallet_record('type', my_id, value)
        opts = RetrieveRecordOptions()
        opts.check_all()
        value_info = await agent1.wallet.non_secrets.get_wallet_record('type', my_id, opts)
        assert value_info['id'] == my_id
        assert value_info['tags'] == {}
        assert value_info['value'] == value
        assert value_info['type'] == 'type'

        tags1 = {
            'tag1': 'val1',
            '~tag2': 'val2'
        }

        await agent1.wallet.non_secrets.update_wallet_record_tags('type', my_id, tags1)
        value_info = await agent1.wallet.non_secrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == tags1

        tags2 = {
            'tag3': 'val3',
        }
        await agent1.wallet.non_secrets.update_wallet_record_tags('type', my_id, tags2)
        value_info = await agent1.wallet.non_secrets.get_wallet_record('type', my_id, opts)
        assert value_info['tags'] == tags2
    finally:
        await agent1.close()


@pytest.mark.asyncio
async def test_record_search(agent1: Agent):
    await agent1.open()
    try:
        id1 = 'id-1-' + uuid.uuid4().hex
        id2 = 'id-2-' + uuid.uuid4().hex
        value1 = 'value-1-' + uuid.uuid4().hex
        value2 = 'value-2-' + uuid.uuid4().hex
        marker_a = 'A-' + uuid.uuid4().hex
        marker_b = 'B-' + uuid.uuid4().hex
        opts = RetrieveRecordOptions()
        opts.check_all()
        tags1 = {
            'tag1': value1,
            '~tag2': '5',
            'marker': marker_a
        }
        tags2 = {
            'tag3': 'val3',
            '~tag4': value2,
            'marker': marker_b
        }
        await agent1.wallet.non_secrets.add_wallet_record('type', id1, 'value1', tags1)
        await agent1.wallet.non_secrets.add_wallet_record('type', id2, 'value2', tags2)

        query = {
            "tag1": value1
        }
        records, total = await agent1.wallet.non_secrets.wallet_search('type', query, opts)
        assert total == 1
        assert 'value-1' in str(records)

        query = {
            "$or": [{"tag1": value1}, {"~tag4": value2}]
        }
        records, total = await agent1.wallet.non_secrets.wallet_search('type', query, opts)
        assert len(records) == 1
        assert total == 2

        records, total = await agent1.wallet.non_secrets.wallet_search('type', query, opts, limit=1000)
        assert len(records) == 2
        assert total == 2

        query = {
            "marker": {"$in": [marker_a, marker_b]}
        }
        records, total = await agent1.wallet.non_secrets.wallet_search('type', query, opts)
        assert 2 == total
    finally:
        await agent1.close()


@pytest.mark.asyncio
async def test_register_schema_in_network(agent2: Agent, default_network: str):
    await agent2.open()
    try:
        seed = '000000000000000000000000Trustee1'
        did, verkey = await agent2.wallet.did.create_and_store_my_did(seed=seed)
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, schema = await agent2.wallet.anoncreds.issuer_create_schema(
            did, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )

        ok, response = await agent2.wallet.ledger.register_schema(default_network, did, schema.body)
        assert ok is True
    finally:
        await agent2.close()


@pytest.mark.asyncio
async def test_register_cred_def_in_network(agent2: Agent, default_network: str):
    await agent2.open()
    try:
        seed = '000000000000000000000000Trustee1'
        did, verkey = await agent2.wallet.did.create_and_store_my_did(seed=seed)
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, schema = await agent2.wallet.anoncreds.issuer_create_schema(
            did, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )

        ok, response = await agent2.wallet.ledger.register_schema(default_network, did, schema.body)
        assert ok is True

        opts = CacheOptions()
        schema_from_ledger = await agent2.wallet.cache.get_schema(default_network, did, schema_id, opts)

        cred_def_id, cred_def = await agent2.wallet.anoncreds.issuer_create_and_store_credential_def(
            did, schema_from_ledger, 'TAG'
        )
        ok, response = await agent2.wallet.ledger.register_cred_def(default_network, did, cred_def)
        assert ok is True
    finally:
        await agent2.close()


@pytest.mark.asyncio
async def test_nym_operations_in_network(agent1: Agent, agent2: Agent, default_network: str):
    await agent1.open()
    await agent2.open()
    steward = agent1.wallet
    actor = agent2.wallet
    try:
        seed = '000000000000000000000000Steward1'
        did_steward, verkey_steward = await steward.did.create_and_store_my_did(seed=seed)
        did_trustee, verkey_trustee = await actor.did.create_and_store_my_did()
        did_common, verkey_common = await actor.did.create_and_store_my_did()

        # Trust Anchor
        ok, response = await steward.ledger.write_nym(
            default_network, did_steward, did_trustee, verkey_trustee, 'Test-Trustee', NYMRole.TRUST_ANCHOR
        )
        assert ok is True
        ok, nym1 = await steward.ledger.read_nym(
            pool_name=default_network, submitter_did=did_steward, target_did=did_trustee
        )
        assert ok is True
        ok, nym2 = await steward.ledger.read_nym(
            pool_name=default_network, submitter_did=None, target_did=did_trustee
        )
        assert ok is True
        assert json.dumps(nym1, sort_keys=True) == json.dumps(nym2, sort_keys=True)
        assert nym1['role'] == str(NYMRole.TRUST_ANCHOR.value[0])

        # Common User
        ok, response = await steward.ledger.write_nym(
            default_network, did_steward, did_common, verkey_common, 'CommonUser', NYMRole.COMMON_USER
        )
        assert ok is True
        ok, nym3 = await steward.ledger.read_nym(
            pool_name=default_network, submitter_did=None, target_did=did_common
        )
        assert ok is True
        assert nym3['role'] is None

        # Reset
        ok, response = await actor.ledger.write_nym(
            default_network, did_common, did_common, verkey_common, 'ResetUser', NYMRole.RESET
        )
        assert ok is True
        ok, nym4 = await steward.ledger.read_nym(
            pool_name=default_network, submitter_did=None, target_did=did_common
        )
        assert ok is True
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_attribute_operations_in_network(agent1: Agent, agent2: Agent, default_network: str):
    await agent1.open()
    await agent2.open()
    steward = agent1.wallet
    actor = agent2.wallet
    try:
        seed = '000000000000000000000000Steward1'
        did_steward, verkey_steward = await steward.did.create_and_store_my_did(seed=seed)
        did_common, verkey_common = await actor.did.create_and_store_my_did()
        ok, response = await steward.ledger.write_nym(
            default_network, did_steward, did_common, verkey_common, 'CommonUser', NYMRole.COMMON_USER
        )
        assert ok is True

        ok, response = await actor.ledger.write_attribute(
            default_network, did_common, did_common, 'attribute', 'value'
        )
        assert ok is True

        ok, response = await steward.ledger.read_attribute(
            default_network, did_steward, did_common, 'attribute'
        )
        assert ok is True
        assert response == 'value'
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_issue_verify_credential_in_network(agent1: Agent, agent2: Agent, agent3: Agent, default_network: str):
    await agent1.open()
    await agent2.open()
    await agent3.open()
    issuer = agent1.wallet
    holder = agent2.wallet
    verifier = agent3.wallet
    try:
        assert 1
    finally:
        await agent1.close()
        await agent2.close()
        await agent3.close()



import uuid

import pytest

from sirius_sdk import Agent
from sirius_sdk.agent.ledger import CredentialDefinition, CacheOptions


@pytest.mark.asyncio
async def test_schema_registration(agent1: Agent):
    await agent1.open()
    try:
        seed = '000000000000000000000000Steward1'
        did, verkey = await agent1.wallet.did.create_and_store_my_did(seed=seed)
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = agent1.ledger('default')

        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did)
        assert ok is True
        assert schema.seq_no > 0

        ok, _ = await ledger.register_schema(schema=anoncred_schema, submitter_did=did)
        assert ok is False

        restored_schema = await ledger.ensure_schema_exists(schema=anoncred_schema, submitter_did=did)
        assert restored_schema is not None
        assert restored_schema == schema

    finally:
        await agent1.close()


@pytest.mark.asyncio
async def test_schema_loading(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    try:
        seed1 = '000000000000000000000000Steward1'
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did(seed=seed1)
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did1, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger1 = agent1.ledger('default')

        ok, schema = await ledger1.register_schema(schema=anoncred_schema, submitter_did=did1)
        assert ok is True
        assert schema.seq_no > 0

        seed2 = '000000000000000000000000Trustee0'
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did(seed=seed2)
        ledger2 = agent2.ledger('default')

        for n in range(5):
            loaded_schema = await ledger2.load_schema(id_=schema.id, submitter_did=did2)
            assert loaded_schema is not None
            assert loaded_schema == schema
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_schema_fetching(agent1: Agent):
    await agent1.open()
    try:
        seed = '000000000000000000000000Steward1'
        did, verkey = await agent1.wallet.did.create_and_store_my_did(seed=seed)
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = agent1.ledger('default')

        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did)
        assert ok is True

        fetches = await ledger.fetch_schemas(name=schema_name)
        assert len(fetches) == 1
        assert fetches[0].issuer_did == did

    finally:
        await agent1.close()


@pytest.mark.asyncio
async def test_register_cred_def(agent1: Agent):
    await agent1.open()
    try:
        seed = '000000000000000000000000Steward1'
        did, verkey = await agent1.wallet.did.create_and_store_my_did(seed=seed)
        schema_name = 'schema_' + uuid.uuid4().hex
        schema_id, anoncred_schema = await agent1.wallet.anoncreds.issuer_create_schema(
            did, schema_name, '1.0', ['attr1', 'attr2', 'attr3']
        )
        ledger = agent1.ledger('default')

        ok, schema = await ledger.register_schema(schema=anoncred_schema, submitter_did=did)
        assert ok is True

        cred_def = CredentialDefinition(tag='Test Tag', schema=schema)
        assert cred_def.body is None
        ok, ledger_cred_def = await ledger.register_cred_def(cred_def=cred_def, submitter_did=did)
        assert ok is True
        assert ledger_cred_def.body is not None
        assert ledger_cred_def.seq_no > 0
        assert ledger_cred_def.submitter_did == did
        my_value = 'my-value-' + uuid.uuid4().hex

        ok, ledger_cred_def2 = await ledger.register_cred_def(
            cred_def=cred_def, submitter_did=did, tags={'my_tag': my_value}
        )
        assert ok is True
        assert ledger_cred_def.body == ledger_cred_def2.body
        assert ledger_cred_def2.seq_no > ledger_cred_def.seq_no

        ser = ledger_cred_def.serialize()
        loaded = CredentialDefinition.deserialize(ser)
        assert loaded.body == ledger_cred_def.body
        assert loaded.seq_no == ledger_cred_def.seq_no
        assert loaded.schema.body == ledger_cred_def.schema.body
        assert loaded.config.serialize() == ledger_cred_def.config.serialize()

        results = await ledger.fetch_cred_defs(schema_id=schema_id)
        assert len(results) == 2
        results = await ledger.fetch_cred_defs(my_tag=my_value)
        assert len(results) == 1

        parts = ledger_cred_def.id.split(':')
        print(str(parts))

        opts = CacheOptions()
        for n in range(3):
            cached_body = await agent1.wallet.cache.get_cred_def('default', did, ledger_cred_def.id, opts)
            assert cached_body == ledger_cred_def.body
            cred_def = await ledger.load_cred_def(ledger_cred_def.id, did)
            assert cred_def.body == cached_body
    finally:
        await agent1.close()
